#!/usr/bin/env python

import inspect

import numpy as np
import networkx as nx

from ..LPU import LPU
from .NDComponent import NDComponent
from ..InputProcessors.BaseInputProcessor import BaseInputProcessor
from ..InputProcessors.ArrayInputProcessor import ArrayInputProcessor
from ..OutputProcessors.OutputRecorder import OutputRecorder
from ..OutputProcessors.FileOutputProcessor import FileOutputProcessor

def test_NDComponent(comp, dt, **kwargs):
    """
    Test an NDComponent.

    Parameters
    ----------
    comp: class or str
          Specifies the NDComponent class to test.
          Either the class itself, or its name.
          If name of the class is used, the path to the module containing
          the class must be specified by the argument `path`.
    dt: float
        time step used for simulation.
    path: str
          Path to the module containing the NDComponent class to test.
    input: InputProcessor or dict
           If input is an instance of the subclass of BaseInputProcessor,
           return input, check there is one 1 uid (can be multiple instances)
           in the InputProcessor, and the length is None.

           If input is a dict,
           should be provided in the following example format:
           {var1: InputProcessor,
            var2: np.ndarray, ...},
           where var1, var2 are variable name to be injected to,
           the values can be InputProcessors or np.ndarray.
           If np.ndarray is provided as input, an ArrayInputProcessor
           will be created for it and uid taking from InputProcessor
           specified from other variables, or 'component' if no InputProcessor
           is used.

           If not specified, no model will run without inputs.
    dur: float
         duration of simulation.
         If not specified, infer from other arguments.
    steps: int
           number of time steps to simulate.
           If not specified, infer from other arguments.
    device: int
            The device ID of the GPU to run stimulation (default: 0).
    print_timing: bool
                  Whether or not to print out the timing of the test (default: False).
    Return
    ------
    output: OurputProcessor.OutputRecorder or str
            If filename not specified, returns an OurputProcessor.OutputRecorder
            object, in which the uid and data of outputs can be accessed by
            `output[var]`.
            If filename is provided, returns the filename of the stored outputs.

    Examples
    --------
    import numpy as np
    from neurokernel.LPU.NDComponents.testing import test_NDComponent
    from neurokernel.LPU.NDComponents.AxonHillockModels.HodgkinHuxley2 import HodgkinHuxley2

    dt = 1e-6
    output = test_NDComponent(HodgkinHuxley2, dt,
                              g_K = 36.0, g_Na = 120.0, g_L = 0.3,
                              E_Na = 50., E_K = -77., E_L = -54.387,
                              input = {'I': np.zeros((int(1e5),1))+10})
    print('spike times:', np.where(output.output['spike_state']['data'])[0]*dt)
    """
    if inspect.isclass(comp):
        if issubclass(comp, NDComponent):
            comp_name = comp.__name__
            extra_comp = comp
        else:
            raise ValueError('comp is a class but not a subclass of NDComponent.')
    elif isinstance(comp, str):
        path = kwargs.get('module_path', [])
        cls = [(name, obj) for name, obj in LPU.import_NDcomponent_from_path(path) is name == comp]
        if len(cls) == 0:
            raise ValueError('Cannot find any {} model from {}'.format(comp, path))
        else:
            comp_name, extra_comp = cls[0]

    accesses = extra_comp.accesses
    updates = extra_comp.updates
    params = extra_comp.params
    extra_params = extra_comp.extra_params
    internals = extra_comp.internals

    def process_input(input, extra_comp):
        """
        extract and create InputProcessors from input

        parameters
        ----------
        input: InputProcessor or dict
               If input is an instance of the subclass of BaseInputProcessor,
               return input, check there is one 1 uid (can be multiple instances)
               in the InputProcessor, and the length is None.

               If input is a dict,
               should be provided in the following example format:
               {var1: InputProcessor,
                var2: np.ndarray, ...},
               where var1, var2 are variable name to be injected to,
               the values can be InputProcessors or np.ndarray.
               If np.ndarray is provided as input, an ArrayInputProcessor
               will be created for it and uid taking from InputProcessor
               specified from other variables, or 'component' if no InputProcessor
               is used.

        extra_comp: class
                    an NDComponent class

        Return
        ------
        input_processors: list
                          a list of input_processors
        uid: str
             uid that can be used to specify the node id
        length: int or None
                If np.ndarray is provided in any of the variables,
                the number of steps will be returned as length.
                Otherwise None is returned.
        """
        if input is None:
            return [], 'component', None

        model_name = extra_comp.__name__
        accesses = set(extra_comp.accesses)
        if isinstance(input, BaseInputProcessor):
            uids = []
            for var in input.variables:
                if var not in accesses:
                    raise ValueError('InputProcessor contains variable {}, not in the accesses field of {} model with accesses = [{}]'.format(var, model_name, ','.join(list(accesses))))
                uids.append(input.variables[var]['uids'])
            uid = set(uids)
            if len(uid > 1):
                raise ValueError('More than 1 uids specified in the InputProcessor')
            uid = list(uid)[0] #just need to return the uid of the component
            return [input], uid, None
        elif isinstance(input, dict):
            length = []
            # check var
            var = set(input.keys())
            assert var.issubset(accesses), 'input contains variable(s) not in the accesses field of {} model with accesses = [{}]'.format(var, model_name, ','.join(list(accesses)))

            # check uid
            input_processors = [v for k, v in input.items() if isinstance(v, BaseInputProcessor)]
            if not len(input_processors):
                uid = 'component'
            else:
                uids = []
                for v in input_processors:
                    for var in v.variables:
                        uids.append(v.variables[var]['uids'])
                uid = set(uids)
                if len(uid > 1):
                    raise ValueError('More than 1 uids specified in the InputProcessor')
                uid = list(uid)[0]
            input_processors = []
            for k, v in input.items():
                if isinstance(v, BaseInputProcessor):
                    input_processors.append(v)
                elif isinstance(v, np.ndarray):
                    assert len(v.shape) == 2, 'Input must be a 2D'
                    input_processors.append(
                        ArrayInputProcessor({k: {'uids': [uid]*v.shape[1], 'data': v}}))
                    length.append(v.shape[0])
                elif isinstance(v, dict) and k == 'spike_state':
                    input_processors.append(
                        ArrayInputProcessor({k: {'uids': [uid]*(v['index'].max()+1), 'data': v}}))
                    length.append(v['time'][-1]+0.5)
            if len(length) == 0:
                length = None
            else:
                length = int(np.max(np.array(length)))
            return input_processors, uid, length

    input = kwargs.get('input', None)
    input_processors, uid, length = process_input(input, extra_comp)

    set_params = {k: v for k, v in kwargs.items() if k in params+extra_params}
    set_internals = {'init{}'.format(k): v for k, v in kwargs.items() if k in internals}

    args = set_params
    args.update(set_internals)
    G = nx.MultiDiGraph()
    G.add_node(uid, name = 'component',
               model = comp_name,
               **args
               )

    dur = kwargs.get('dur', None)
    steps = kwargs.get('steps', None)

    if steps is None:
        if dur is not None:
            steps = int(np.floor(dur/dt))
        else:
            if input is not None and length is not None:
                steps = length
                dur = steps*dt
    if dur is None:
        if steps is not None:
            dur = steps*dt
        else:
            if input is not None and length is not None:
                steps = length
                dur = steps*dt

    output_processor = OutputRecorder([(var, [uid]) for var in updates])

    print('Testing component {} for duration {} at dt = {} and a total number of {} steps...'.format(
                            comp_name, dur, dt, steps))

    lpu = LPU(dt, 'obj', {'graph': G, 'kwargs': {'class_key': 'model'}},
              device = kwargs.get('device', 0),
              id = 'comp_test', input_processors = input_processors,
              output_processors = [output_processor],
              debug = True,
              manager = False,
              print_timing = kwargs.get('print_timing', False),
              extra_comps = [extra_comp])

    lpu.run(steps = steps)

    return output_processor.get_output(uids = uid)[uid]



from .LPU import LPU

class Transform(Object):
    def __init__(self, LPUobj):
        assert(isinstance(LPUobj, LPU))
        self._obj_ids = None
        self._LPU_obj = LPUobj
        self._kernel = None
        self._need_to_compile_kernel = True
        self._need_to_generate_kernel = True
        self._variable_name = None
        self._custom_kernel = False
        self._kernel_string = None

    def _compile_kernel(self):
        pass

    def _gen_kernel_code(self):
        pass

    @property
    def compiled_kernel(self):
        if self._need_to_compile_kernel:
            self._compile_kernel()
        return self._kernel

    @property
    def variable_name(self):
        return self._variable_name

    @variable_name.setter
    def variable_name(self, value):
        assert(self.LPU_obj.is_variable(value)):
        self ._variable_name = value

    @property
    def obj_ids(self):
        return self._obj_ids

    @obj_ids.setter
    def obj_ids(self, value):
        assert(self._variable_name)
        self._obj_ids = value

    @property
    def LPU_obj(self):
        return self._LPU_obj

    @property
    def kernel(self):
        if self._need_to_generate_kernel:
            self._gen_kernel_code()
        return self._kernel_string

    @kernel.setter
    def kernel(self, value):
        assert(isinstance(value, string))
        self._custom_kernel = True
        self._need_to_compile_kernel = True
        self._need_to_generate_kernel = False
        self._kernel_string = value



import os
import fnmatch

__all__ = []

NDC_dir = os.path.dirname(__file__)
for root, dirnames, filenames in os.walk(NDC_dir):
    mod_imp = False
    for f in fnmatch.filter(filenames,"*.py"):
        if '__init__'!=f[:8]:
            if root != NDC_dir and not mod_imp:
                __path__.append(root)
                mod_imp = True
            __all__.append(f[:-3])



from neurokernel.LPU.NDComponents.SynapseModels.BaseSynapseModel import *

#This class assumes a single pre synaptic connection per component instance
class PowerGPotGPot(BaseSynapseModel):
    accesses = ['V']
    updates = ['g']
    params = ['threshold', 'slope', 'power', 'saturation']
    extra_params = []
    internals = OrderedDict([])

    @property
    def maximum_dt_allowed(self):
        return self.dt

    def get_update_template(self):
        template = """
__global__ void update(int num_comps, %(dt)s dt, int steps,
                       %(input_V)s* g_V, %(param_threshold)s* g_threshold,
                       %(param_slope)s* g_slope, %(param_power)s* g_power,
                       %(param_saturation)s* g_saturation,
                       %(update_g)s* g_g)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    %(dt)s ddt = dt*1000.; // s to ms
    %(input_V)s V;
    %(param_threshold)s threshold;
    %(param_slope)s slope;
    %(param_power)s power;
    %(param_saturation)s saturation;

    for(int i = tid; i < num_comps; i += total_threads)
    {
        V = g_V[i];
        threshold = g_threshold[i];
        slope = g_slope[i];
        power = g_power[i];
        saturation = g_saturation[i];

        g_g[i] = fmin%(fletter)s(saturation,
                    slope*pow%(fletter)s(fmax(0.0,V-threshold),power));
    }
}
"""
        return template


if __name__ == '__main__':
    import argparse
    import itertools
    import networkx as nx
    from neurokernel.tools.logging import setup_logger
    import neurokernel.core_gpu as core

    from neurokernel.LPU.LPU import LPU

    from neurokernel.LPU.InputProcessors.RampInputProcessor import RampInputProcessor
    from neurokernel.LPU.OutputProcessors.FileOutputProcessor import FileOutputProcessor

    import neurokernel.mpi_relaunch

    dt = 1e-4
    dur = 1.0
    steps = int(dur/dt)

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', default=False,
                        dest='debug', action='store_true',
                        help='Write connectivity structures and inter-LPU routed data in debug folder')
    parser.add_argument('-l', '--log', default='none', type=str,
                        help='Log output to screen [file, screen, both, or none; default:none]')
    parser.add_argument('-s', '--steps', default=steps, type=int,
                        help='Number of steps [default: %s]' % steps)
    parser.add_argument('-g', '--gpu_dev', default=0, type=int,
                        help='GPU device number [default: 0]')
    args = parser.parse_args()

    file_name = None
    screen = False
    if args.log.lower() in ['file', 'both']:
        file_name = 'neurokernel.log'
    if args.log.lower() in ['screen', 'both']:
        screen = True
    logger = setup_logger(file_name=file_name, screen=screen)

    man = core.Manager()

    G = nx.MultiDiGraph()

    G.add_node('synapse0', **{
               'class': 'PowerGPotGPot',
               'name': 'PowerGPotGPot',
               'gmax': 0.4,
               'threshold': -55.0,
               'slope': 0.02,
               'power': 1.0,
               'saturation': 0.4,
               'reverse': 0.0
               })

    comp_dict, conns = LPU.graph_to_dicts(G)

    fl_input_processor = RampInputProcessor('V', ['synapse0'], 0.0, 1.0, -70.0, -30.0)
    fl_output_processor = FileOutputProcessor([('g', None)], 'new_output.h5', sample_interval=1)

    man.add(LPU, 'ge', dt, comp_dict, conns,
            device=args.gpu_dev, input_processors = [fl_input_processor],
            output_processors = [fl_output_processor], debug=args.debug)

    man.spawn()
    man.start(steps=args.steps)
    man.wait()



#!/usr/bin/env python
from abc import ABCMeta, abstractmethod, abstractproperty
from future.utils import with_metaclass
from neurokernel.LPU.NDComponents.NDComponent import NDComponent
from neurokernel.LPU.utils.simpleio import *

from collections import OrderedDict

import numpy as np

import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype
import pycuda.driver as cuda
from pycuda.compiler import SourceModule


class BaseMembraneModel(with_metaclass(ABCMeta, NDComponent)):
    # __metaclass__ = ABCMeta

    accesses = ['I']
    updates = ['V']
    extra_params = []
    params = []
    internals = OrderedDict([('V', -65.)])

    def __init__(self, params_dict, access_buffers, dt,
                 debug=False, LPU_id=None, cuda_verbose=True):
        if cuda_verbose:
            self.compile_options = ['--ptxas-options=-v']
        else:
            self.compile_options = []

        self.num_comps = params_dict[self.params[0]].size
        self.params_dict = params_dict
        self.access_buffers = access_buffers

        self.debug = debug
        self.LPU_id = LPU_id
        self.dtype = params_dict[self.params[0]].dtype

        self.dt = np.double(dt)

        self.internal_states = {
            c: garray.zeros(self.num_comps, dtype = self.dtype)+self.internals[c] \
            for c in self.internals}

        self.inputs = {
            k: garray.empty(self.num_comps, dtype = self.access_buffers[k].dtype)\
            for k in self.accesses}

        dtypes = {'dt': self.dtype}
        dtypes.update({'input_{}'.format(k): self.inputs[k].dtype for k in self.accesses})
        dtypes.update({'param_{}'.format(k): self.params_dict[k].dtype for k in self.params})
        dtypes.update({'internal_{}'.format(k): self.internal_states[k].dtype for k in self.internals})
        dtypes.update({'update_{}'.format(k): self.dtype for k in self.updates})
        self.update_func = self.get_update_func(dtypes)

    def pre_run(self, update_pointers):
        self.add_initializer('initV', 'V', update_pointers)

    def run_step(self, update_pointers, st=None):
        for k in self.inputs:
            self.sum_in_variable(k, self.inputs[k], st=st)

        self.update_func.prepared_async_call(
            self.update_func.grid, self.update_func.block, st,
            self.num_comps, self.internal_dt, self.internal_steps,
            *[self.inputs[k].gpudata for k in self.accesses]+\
            [self.params_dict[k].gpudata for k in self.params]+\
            [self.internal_states[k].gpudata for k in self.internals]+\
            [update_pointers[k] for k in self.updates])

    def get_update_template(self):
        template = ''
        return template

    def get_update_func(self, dtypes):
        type_dict = {k: dtype_to_ctype(dtypes[k]) for k in dtypes}
        type_dict.update({'fletter': 'f' if type_dict['param_{}'.format(self.params[0])] == 'float' else ''})
        mod = SourceModule(self.get_update_template() % type_dict,
                           options=self.compile_options)
        func = mod.get_function("update")
        func.prepare('i'+np.dtype(dtypes['dt']).char+'i'+'P'*(len(type_dict)-2))
        func.block = (128,1,1)
        func.grid = (min(6 * cuda.Context.get_device().MULTIPROCESSOR_COUNT,
                         (self.num_comps-1) // 128 + 1), 1)
        return func



from neurokernel.LPU.NDComponents.AxonHillockModels.BaseAxonHillockModel import *

class LeakyIAF(BaseAxonHillockModel):
    # updates are the output states of the model
    updates = ['spike_state', # (bool)
               'V' # Membrane Potential (mV)
              ]
    # accesses are the input variables of the model
    accesses = ['I'] # (\mu A/cm^2 )
    # params are the parameters of the model that needs to be defined
    # during specification of the model
    params = ['resting_potential', # (mV)
              'threshold', # Firing Threshold (mV)
              'reset_potential', # Potential to be reset to after a spike (mV)
              'capacitance', # (\mu F/cm^2)
              'resistance' # (k\Omega cm.^2)
              ]
    extra_params = ['initV']
    # internals are the variables used to store internal states of the model,
    # and are ordered dict whose keys are the variables and value are the initial values.
    internals = OrderedDict([('V', 0.0)]) # Membrane Potential (mV)

    @property
    def maximum_dt_allowed(self):
        return 1e-4

    def pre_run(self, update_pointers):
        super(LeakyIAF, self).pre_run(update_pointers)
        if 'initV' not in self.params_dict:
            self.add_initializer('resting_potential', 'V', update_pointers)

    def get_update_template(self):
        # need to update the CUDA kernel to reflect the equations of the model
        # the argument of the function must in the following order:
        # 1. int num_comps
        # 2. %(dt)s dt,
        # 3. int nsteps,
        # 4. all variables in accesses according to the order in accesses
        # 5. all variables in params according to the order in params
        # 6. all variables in internals according to the order in internals
        # 7. all variables in updates according to the order in updates
        template = """
__global__ void update(int num_comps,
               %(dt)s dt,
               int nsteps,
               %(input_I)s* g_I, // accesses
               %(param_resting_potential)s* g_resting_potential, // params
               %(param_threshold)s* g_threshold,
               %(param_reset_potential)s* g_reset_potential,
               %(param_capacitance)s* g_capacitance,
               %(param_resistance)s* g_resistance,
               %(internal_V)s* g_internalV, // internals
               %(update_spike_state)s* g_spike_state, //updates
               %(update_V)s* g_V)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    // instantiate variables
    %(dt)s ddt = dt*1000.; // s to ms
    %(update_V)s V;
    %(input_I)s I;
    %(update_spike_state)s spike;
    %(param_resting_potential)s resting_potential;
    %(param_threshold)s threshold;
    %(param_reset_potential)s reset_potential;
    %(param_capacitance)s capacitance;
    %(param_resistance)s resistance;
    %(dt)s bh;

    // no need to change this for loop
    for(int i = tid; i < num_comps; i += total_threads)
    {
        // load the data from global memory
        spike = 0;
        V = g_internalV[i];
        I = g_I[i];
        capacitance = g_capacitance[i];
        resting_potential = g_resting_potential[i];
        threshold = g_threshold[i];
        resistance = g_resistance[i];
        reset_potential = g_reset_potential[i];

        // update according to equations of the model
        bh = exp%(fletter)s(-ddt/(capacitance*resistance));

        for (int j = 0; j < nsteps; ++j)
        {
            V = V*bh + (resistance*I+resting_potential)*(1.0 - bh);
            if (V >= threshold)
            {
                V = reset_potential;
                spike = 1.0;
            }
        }

        // write local updated states back to global memory
        g_V[i] = V;
        g_internalV[i] = V;
        g_spike_state[i] = spike;
    }
}
        """
        return template



if __name__ == '__main__':
    import argparse
    import itertools
    import networkx as nx
    from neurokernel.tools.logging import setup_logger
    import neurokernel.core_gpu as core

    from neurokernel.LPU.LPU import LPU

    from neurokernel.LPU.InputProcessors.FileInputProcessor import FileInputProcessor
    from neurokernel.LPU.InputProcessors.StepInputProcessor import StepInputProcessor
    from neurokernel.LPU.OutputProcessors.FileOutputProcessor import FileOutputProcessor

    import neurokernel.mpi_relaunch

    dt = 1e-4
    dur = 1.0
    steps = int(dur/dt)

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', default=False,
                        dest='debug', action='store_true',
                        help='Write connectivity structures and inter-LPU routed data in debug folder')
    parser.add_argument('-l', '--log', default='none', type=str,
                        help='Log output to screen [file, screen, both, or none; default:none]')
    parser.add_argument('-s', '--steps', default=steps, type=int,
                        help='Number of steps [default: %s]' % steps)
    parser.add_argument('-g', '--gpu_dev', default=0, type=int,
                        help='GPU device number [default: 0]')
    args = parser.parse_args()

    file_name = None
    screen = False
    if args.log.lower() in ['file', 'both']:
        file_name = 'neurokernel.log'
    if args.log.lower() in ['screen', 'both']:
        screen = True
    logger = setup_logger(file_name=file_name, screen=screen)

    man = core.Manager()

    G = nx.MultiDiGraph()

    # specify the node
    G.add_node('neuron0', **{
               'class': 'LeakyIAF',
               'name': 'LeakyIAF',
               'resting_potential': -70.0,
               'reset_potential': -70.0,
               'threshold': -45.0, #
               'capacitance': 2.0, # in mS
               'resistance': 10.0, # in Ohm
               })

    comp_dict, conns = LPU.graph_to_dicts(G)

    # use a input processor that present a step current (I) input to 'neuron0'
    # the step is from 0.2 to 0.8 and the step height is 10.0
    fl_input_processor = StepInputProcessor('I', ['neuron0'], 10.0, 0.2, 0.8)
    # output processor to record 'spike_state' and 'V' to hdf5 file 'new_output.h5',
    # with a sampling interval of 1 run step.
    fl_output_processor = FileOutputProcessor([('spike_state', None),('V', None)], 'new_output.h5', sample_interval=1)

    man.add(LPU, 'ge', dt, comp_dict, conns,
            device=args.gpu_dev, input_processors = [fl_input_processor],
            output_processors = [fl_output_processor], debug=args.debug)

    man.spawn()
    man.start(steps=args.steps)
    man.wait()

    import h5py
    import matplotlib
    matplotlib.use('PS')
    import matplotlib.pyplot as plt

    f = h5py.File('new_output.h5')
    t = np.arange(0, args.steps)*dt

    plt.figure()
    plt.plot(t,list(f['V'].values())[0])
    plt.xlabel('time, [s]')
    plt.ylabel('Voltage, [mV]')
    plt.title('Leaky Integrate-and-Fire Neuron')
    plt.savefig('lif.png',dpi=300)



from pycuda.compiler import SourceModule
import pycuda.gpuarray as garray
import numpy as np

size_of_curandStateXORWOW = 12

def curand_setup(num_threads, seed):
    """
    Setup curand seed
    """
    func = get_curand_int_func()
    grid = ( (int(num_threads)-1)//128 + 1,1)
    block = (128,1,1)

    # curandStateXORWOW is a struct of size 12 4bytes
    state = garray.empty(num_threads*size_of_curandStateXORWOW, np.int32)

    func.prepared_call( grid, block, state.gpudata, num_threads, seed)
    return state


def get_curand_int_func():
    code = """
#include "curand_kernel.h"
extern "C" {
__global__ void
rand_setup(curandStateXORWOW_t* state, int size, unsigned long long seed)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = blockDim.x * gridDim.x;

    for(int i = tid; i < size; i+=total_threads)
    {
        curand_init(seed, i, 0, &state[i]);
    }
}
}
    """
    mod = SourceModule(code, no_extern_c = True)
    func = mod.get_function("rand_setup")
    func.prepare('PiL')#[np.intp, np.int32, np.uint64])
    return func



#!/usr/bin/env python


import warnings
from abc import ABCMeta, abstractmethod, abstractproperty
import os.path
import numpy as np
from future.utils import with_metaclass

import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype
import pycuda.driver as cuda
from pycuda.compiler import SourceModule

from neurokernel.LPU.utils.simpleio import *

class NDComponent(with_metaclass(ABCMeta, object)):
    # __metaclass__ = ABCMeta

    accesses = []
    updates = []

    @abstractmethod
    def __init__(self, params_dict, access_buffers, dt, debug=False,
                 LPU_id=None, cuda_verbose=False):
        pass

    @abstractmethod
    def run_step(self, update_pointers):
        pass

    @property
    @abstractmethod
    def maximum_dt_allowed(self):
        pass

    @property
    def internal_steps(self):
        if self.dt > self.maximum_dt_allowed:
            div = self.dt/self.maximum_dt_allowed
            if np.abs(div - np.round(div)) < 1e-5:
                return int(np.round(div))
            else:
                return int(np.ceil(div))
            #raise ValueError('Simulation time step dt larger than maximum allowed dt of model {}'.format(type(self)))
        else:
            return 1

    @property
    def internal_dt(self):
        return self.dt/self.internal_steps

    def pre_run(self, update_pointers):
        pass

    def post_run(self):
        '''
        This method will be called at the end of the simulation.
        '''
        pass

    def add_initializer(self, var_a, var_b, update_pointers):
        if var_a in self.params_dict:
            if var_b in self.internal_states:
                cuda.memcpy_dtod(self.internal_states[var_b].gpudata,
                                    self.params_dict[var_a].gpudata,
                                    self.params_dict[var_a].nbytes)
            if var_b in update_pointers:
                cuda.memcpy_dtod(int(update_pointers[var_b]),
                                    self.params_dict[var_a].gpudata,
                                    self.params_dict[var_a].nbytes)

    def sum_in_variable(self, var, garr, st=None):
        try:
            a = self.sum_kernel
        except AttributeError:
            self.sum_kernel = self.__get_sum_kernel(garr.size, garr.dtype)
        self.sum_kernel.prepared_async_call(
            self.__grid_sum, self.__block_sum, st,
            garr.gpudata,                                          #P
            self.params_dict['conn_data'][var]['delay'].gpudata,   #P
            self.params_dict['cumpre'][var].gpudata,               #P
            self.params_dict['npre'][var].gpudata,                 #P
            self.params_dict['pre'][var].gpudata,                  #P
            self.access_buffers[var].gpudata,                      #P
            self.access_buffers[var].ld,                           #i
            self.access_buffers[var].current,                      #i
            self.access_buffers[var].buffer_length)                #i

    def __get_sum_kernel(self, num_comps, dtype=np.double):
        template = """
        #define NUM_COMPS %(num_comps)d

        __global__ void sum_input(%(type)s* res, int* delay, int* cumpre,
                                  int* npre, int* pre, %(type)s* pre_buffer,
                                  int ld, int current, int buffer_length)
        {
            // must use block size (32, 32, 1)
            int tidx = threadIdx.x;
            int tidy = threadIdx.y;
            int bid = blockIdx.x;

            int comp;

            __shared__ int num_pre[32];
            __shared__ int pre_start[32];
            __shared__ double input[32][33];

            if(tidy == 0)
            {
                comp = bid * 32 + tidx;
                if(comp < NUM_COMPS)
                {
                    num_pre[tidx] = npre[comp];
                }
            } else if(tidy == 1)
            {
                comp = bid * 32 + tidx;
                if(comp < NUM_COMPS)
                {
                    pre_start[tidx] = cumpre[comp];
                }
            }

            input[tidy][tidx] = 0.0;

            __syncthreads();

            comp = bid * 32 + tidy ;
            if(comp < NUM_COMPS){
               int dl;
               int col;
               int n_pre = num_pre[tidy];
               int start = pre_start[tidy];

               for(int i = tidx; i < n_pre; i += 32)
               {
                   dl = delay[start+i];
                   col = current - dl;
                   if(col < 0)
                   {
                     col = buffer_length + col;
                   }

                   input[tidy][tidx] += pre_buffer[col*ld + pre[start+i]];
               }
            }
            __syncthreads();

            if(tidy < 8)
            {
                input[tidx][tidy] += input[tidx][tidy + 8];
                input[tidx][tidy] += input[tidx][tidy + 16];
                input[tidx][tidy] += input[tidx][tidy + 24];
            }

            __syncthreads();

            if(tidy < 4)
            {
                input[tidx][tidy] += input[tidx][tidy + 4];
           }

            __syncthreads();

            if(tidy < 2)
            {
                input[tidx][tidy] += input[tidx][tidy + 2];
            }

            __syncthreads();

            if(tidy == 0)
            {
                input[tidx][0] += input[tidx][1];
                comp = bid*32+tidx;
                if(comp < NUM_COMPS)
                {
                    res[comp] = input[tidx][0];
                }
            }

        }
        //can be improved
        """
        mod = SourceModule(template % {"num_comps": num_comps,
                                       "type": dtype_to_ctype(dtype)},
                           options=self.compile_options)
        func = mod.get_function("sum_input")
        func.prepare('PPPPPPiii')
        self.__block_sum = (32, 32, 1)
        self.__grid_sum = ((num_comps - 1) // 32 + 1, 1)
        return func



import numpy as np

from .BaseInputProcessor import BaseInputProcessor


class RampInputProcessor(BaseInputProcessor):

    def __init__(self, variable, uids, start_time, duration, start_value, stop_value):
        super(RampInputProcessor, self).__init__([(variable, uids)],
                                                 mode = 0,
                                                 memory_mode = 'gpu')
        self.duration = duration
        self.start_time = start_time
        self.start_value = start_value
        self.stop_value = stop_value
        self.var = variable
        self.num = len(uids)

    def update_input(self):
        current_time = self.LPU_obj.time
        if current_time <= self.start_time:
            val = self.start_value
        elif current_time >= self.start_time + self.duration:
            val = self.stop_value
        else:
            val = (self.stop_value - self.start_value) / self.duration *\
                  (current_time - self.start_time) \
                + self.start_value
        self.variables[self.var]['input'].fill(val) #* np.ones(self.num, dtype=self.dtypes[self.var])

    def is_input_available(self):
        return True

    def post_run(self):
        pass




from collections import OrderedDict
from datetime import datetime

import h5py
import numpy as np

import pycuda.driver as cuda
import pycuda.gpuarray as garray
from neurokernel.LPU.InputProcessors.BaseInputProcessor import BaseInputProcessor


class ArrayInputProcessor(BaseInputProcessor):
    """
    Parameters
    ----------
    inputs: dict
            A dictionary with the following structure:
            {var1: {'uids': ['comp1', 'comp2' ... ], 'data': np.ndarray},
             var2: {'uids': ['comp3', 'comp4' ... ], 'data': np.ndarray},
             ...}
            where in each key/value pair, the key specifies the varialbe to be
            inject input to, value['uids'] is a list that specifies the uids of
            components to be injected to, and value['data'] is an np.ndarray
            of shape (T, N) that specifies the input to each of the N components
            at every time step for a total of T steps, and the ordering the data in
            each column should be the same as the order in value['uids'].
            For 'spike_state', in addition to provide 0 (no spike) or 1 (spike)
            for each time step, spike time and the index of neuron that spiked
            can be provided by value['data']['time'] and value['data']['index'],
            respectively. See examples for more details.

            The input can also be programmatically loaded by other methods.
    mode:     int
              mode in BaseInputProcessor
              0: default, when the file does not cover the entire simulation,
                 inputs default to 0.
              1: when the file does not cover the entire simulation,
                 inputs defaults to the last state.
    cache_length: int
                  length (number of steps) of cache on GPU memory to preload
                  inputs.

    Examples
    --------
    Input h5d must have the following format:
    1. For 'spike_state', there are two options. First,
        {'spike_state':
            {'uids': numpy array of dtype 'S',
                             contains the uids of the nodes recorded,
                             the order of this array determines the index below,
            'data': {
                'time': numpy array of np.float32/np.double,
                      contains a **monotonically** increasing time stamps
                      of spikes from all nodes,
                'index': numpy array of np.int32
                       contains the index of every entry of spike time
                       in 'spike_state/data/time'}.
            }
        }
        loading a output file generated by FileOutputProcessor into a Dictionary
        of the above structure complies with this requirement.

       A second option is to have it as the same format as in 2 below, but have
       only float/double value of 0 (no spike at the time step) or 1 (spike)

    2. For other variables, e.g., 'I':
        {'I':
            {'uids': numpy array of dtype 'S',
                  contains the uids of the nodes recorded,
                  the order of this array determines which column each node
                  is receives as input in the 'I/data' array.,
             'data': numpy array of dtype np.float32/np.double,
                   contains the value of inputs injected to the nodes.
            }
        }
    """
    def __init__(self, inputs = {}, mode = 0, cache_length = 1000,
                 input_file = None, input_interval = 1,
                 sensory_file = None, sensory_interval = 1):
        super(ArrayInputProcessor, self).__init__([], mode, memory_mode = 'gpu',
                                                  input_file = input_file,
                                                  input_interval = input_interval,
                                                  sensory_file = sensory_file,
                                                  sensory_interval = sensory_interval)
        self.cache_length = cache_length
        self.data = {}
        self.add_inputs(inputs)

    def add_inputs(self, inputs):
        for var, g in inputs.items():
            self.add_input(var, g['uids'], g['data'])

    def add_input(self, var, uids, data):
        """
        Add inputs to the input processor

        Parameters
        ----------
        var: str
             the variable to which input will be injected to the components
        uids: list
              the uids of components to inject to
        data: ndarray or dict
              If `variable` is 'spike_state', then data can be in spike event
              format, as a dictionary {'time': np.ndarray, 'index', np.ndarray},\
              specifying the time and index of spikes.
              For all variables, data can be in stepwise input format,
              as a ndarray of shape (T, N)
              that specifies the input to each of the N components
              at every time step for a total of T steps.
              It is expected that if 'spike_state' input is added more than once,
              they must all be of one form, either in spike event format,
              or in stepwise input format.
        """
        if var in self.variables:
            if var == 'spike_state':
                if isinstance(data, dict):
                    assert issubclass(data['index'].dtype.type, np.integer), \
                           'spike index array must be integer dtype'
                    self.data[var]['time'] = np.concatenate(
                            self.data[var]['time'], data['time'])
                    self.data[var]['index'] = np.concatenate(
                            self.data[var]['index'],
                            data['index']+len(self.variables[var]['uids']))
                else:
                    self.data[var] = np.hstack((self.data[var], data))
            else:
                self.data[var] = np.hstack((self.data[var], data))
        else:
            self.data[var] = data
            if var == 'spike_state':
                if isinstance(data, dict):
                    self.spike_state_format = 'event'
                else:
                    self.spike_state_format = 'stepwise'
        self.add_variables([(var, uids)])

    def pre_run(self):
        self.dsets = {}
        self.cache = {}
        self.counts = {}
        self.end_of_var_in_array = {}
        self.block_total = {}
        self.end_of_var = {}
        self.last_read_index = {}

        if 'spike_state' in self.variables:
            if self.spike_state_format == 'event':
                sort_order = np.argsort(self.data['spike_state']['time'])
                self.data['spike_state']['time'] = self.data['spike_state']['time'][sort_order]
                self.data['spike_state']['index'] = self.data['spike_state']['index'][sort_order]

        for var in self.variables:
            data = self.data[var]
            uids = self.variables[var]['uids']
            self.block_total[var] = 0
            if var == 'spike_state':
                if self.spike_state_format == 'event':
                    self.dsets[var] = {'index': data['index'],
                                       'time': data['time']}
                else:
                    self.dsets[var] = data
                self.cache[var] = garray.empty((self.cache_length, len(self.variables[var]['uids'])),
                                               self.variables[var]['input'].dtype)
            else:
                self.dsets[var] = data
                self.cache[var] = garray.empty((self.cache_length, len(self.variables[var]['uids'])),
                                               self.variables[var]['input'].dtype)
            self.last_read_index[var] = 0
            self.counts[var] = 0
            self.end_of_var[var] = False
            self.end_of_var_in_array[var] = False
        # self.end_of_input = False

    def update_input(self):
        for var, dset in self.dsets.items():
            if self.counts[var] == self.block_total[var]:
                if not self.end_of_var_in_array[var]:
                    self.read_to_cache(var)

            if not self.end_of_var[var]:
                cuda.memcpy_dtod(self.variables[var]['input'].gpudata,
                                 int(self.cache[var].gpudata) + \
                                 self.cache[var].shape[1]*self.counts[var]*self.cache[var].dtype.itemsize,
                                 self.variables[var]['input'].nbytes)
                self.counts[var] += 1
            else:
                if not self.end_of_input:
                    self.variables[var]['input'].fill(0)

            if self.counts[var] == self.block_total[var]:
                if self.end_of_var_in_array[var]:
                    self.end_of_var[var] = True

    def read_to_cache(self, var):
        if self.end_of_var_in_array[var]:
            # self.cache[var].fill(0)
            self.counts[var] = 0
            self.block_total[var] = self.cache_length
            return

        if var == 'spike_state':
            if self.spike_state_format == 'event':
                self.read_spike_times(var)
            else:
                self.read_block_input(var)
        else:
            self.read_block_input(var)
        self.counts[var] = 0

    def read_block_input(self, var):
        tmp = self.dsets[var][self.last_read_index[var]:self.last_read_index[var]+self.cache_length,:]
        self.last_read_index[var] += tmp.shape[0]
        if self.last_read_index[var] == self.dsets[var].shape[0]:
            self.end_of_var_in_array[var] = True
        self.block_total[var] = tmp.shape[0]
        if tmp.shape[0] < self.cache_length:
            tmp1 = np.zeros(self.cache[var].shape, self.cache[var].dtype)
            tmp1[:tmp.shape[0]] = tmp
            tmp = tmp1
        self.cache[var].set(tmp)

    def read_spike_times(self, var):
        current_time = self._LPU_obj.time
        next_time = current_time + self.cache_length*self.sim_dt
        spike_times = []
        spike_index = []
        i = 0
        while True:
            tmp = self.dsets[var]['time'][self.last_read_index[var]:self.last_read_index[var]+10000]
            if tmp[0] + self.sim_dt/100 >= next_time:
                break
            last_spike = tmp.shape[0] - np.argmax(tmp[::-1] + self.sim_dt/100 < next_time)
            index = self.dsets[var]['index'][self.last_read_index[var]:self.last_read_index[var]+last_spike]
            spike_times.append(tmp[:last_spike])
            spike_index.append(index[:last_spike])
            self.last_read_index[var] += last_spike
            if self.last_read_index[var] == self.dsets[var]['time'].shape[0]:
                self.end_of_var_in_array[var] = True
                break
            if last_spike < tmp.shape[0]:
                break
        if len(spike_times):
            spike_times = np.concatenate(spike_times) - current_time
            spike_time_ids = np.floor(spike_times/self.sim_dt+np.finfo(spike_times.dtype).eps*(1/self.sim_dt)).astype(np.int32)
            spike_index = np.concatenate(spike_index)
            tmp = np.zeros(self.cache[var].shape, self.cache[var].dtype)
            self.block_total[var] = self.cache[var].shape[0]
            np.add.at(tmp, (spike_time_ids, spike_index), 1)
            self.cache[var].set(tmp)
        else:
            self.cache[var].fill(0)
            self.block_total[var] = self.cache[var].shape[0]

    @property
    def end_of_input(self):
        return all(self.end_of_var.values())

    def is_input_available(self):
        return not self.end_of_input

    def get_input(self, var = None, uids = None, dt = 1, start_time = 0):
        """
        retrieve inputs by variable name `var` and uids

        Parameters
        ----------
        var: str
             Name of the variable to retrieve.
             If None, all variables associated with a uid will be retrieved.
        uids: str or list of str
              uids of the component to retrieve
              If not specified (None), all uids associated with the `var`
              will be retrieved.

        Returns
        -------
        input: dict or OrderedDict
                If uids is a list or tuple, returns a OrderedDict, Otherwise a dict,
                with inputs keyed by uid, and values are either the data/spike_times
                of the component, or a dict keyed by variable name and data/spike_times
                in the value.
        """
        if var is None and uids is None:
            return self.data
        elif var is not None and uids is None:
            return self._get_input_by_var(var, dt = dt, start_time = start_time)
        elif var is None and uids is not None:
            return self._get_input_by_uids(uids, dt = dt, start_time = start_time)
        else:
            return self._get_input_by_var_and_uids(var, uids, dt = dt, start_time = start_time)

    def _get_input_by_var(self, var, dt = 1, start_time = 0):
        uids = self.variables[var]['uids']
        data = self.data[var]
        if var == 'spike_state' and self.spike_state_format == 'event':
            input = {uid: {'data': data['time'][data['index']==i]} for i, uid in enumerate(uids)}
        else:
            t = np.arange(0, data.shape[0])*dt + start_time
            input = {uid: {'time': t, 'data': data[:,i].copy()} for i, uid in enumerate(uids)}
        return input

    def _get_input_by_uid(self, uid, dt = 1, start_time = 0):
        input = {}
        for var in self.data:
            try:
                index = self.variables[var]['uids'].index(uid)
            except ValueError:
                pass
            else:
                if var == 'spike_state' and self.spike_state_format == 'event':
                    input[var] = self.data[var]['time'][self.data[var]['index'] == index]
                else:
                    t = np.arange(0, self.data[var].shape[0])*dt + start_time
                    input[var] = {'time': t, 'data': self.data[var][:,index].copy()}
        return input

    def _get_input_by_uids(self, uids, dt = 1, start_time = 0):
        if isinstance(uids, str):
            return {uids: self._get_input_by_uid(uids, dt = dt, start_time = start_time)}
        elif isinstance(uids, (list, tuple)):
            return OrderedDict([(uid, self._get_input_by_uid(uid, dt = dt, start_time = start_time)) for uid in uids])
        elif isinstance(uids, set):
            return {uid: self._get_input_by_uid(uid, dt = dt, start_time = start_time) for uid in uids}

    def _get_input_by_var_and_uid(self, var, uid, dt = 1, start_time = 0):
        input = None
        try:
            index = self.variables[var]['uids'].index(uid)
        except ValueError:
            pass
        else:
            if var == 'spike_state' and self.spike_state_format == 'event':
                input = {'data': self.data[var]['time'][self.data[var]['index'] == index]}
            else:
                t = np.arange(0, self.data[var].shape[0])*dt + start_time
                input = {'time': t, 'data': self.data[var][:,index].copy()}
        return input

    def _get_input_by_var_and_uids(self, var, uids):
        if isinstance(uids, str):
            return {uids: self._get_input_by_var_and_uid(var, uids)}
        elif isinstance(uids, (list, tuple)):
            return OrderedDict([(uid, self._get_input_by_var_and_uid(var, uid)) for uid in uids])
        elif isinstance(uids, set):
            return {uid: self._get_input_by_var_and_uid(var, uid) for uid in uids}

    def to_file(self, filename):
        """
        Store inputs to file with name `filename`,
        compatible to be used by `FileInputProcessor`

        parameters
        ----------
        filename: str
                  name of the file to store the results
        """
        with h5py.File(filename, 'w') as h5file:
            h5file.create_dataset('metadata', (), 'i')
            h5file['metadata'].attrs['start_time'] = 0.0
            h5file['metadata'].attrs['sample_interval'] = self.input_interval
            #h5file['metadata'].attrs['dt'] = self.dt
            h5file['metadata'].attrs[
                'DateCreated'] = datetime.now().isoformat()
            for var, d in self.variables.items():
                if var == 'spike_state' and self.spike_state_format == 'event':
                    h5file.create_dataset(var + '/data/index',
                                          dtype = np.int32,
                                          maxshape = (None,),
                                          data = self.data[var]['index'])
                    h5file.create_dataset(var + '/data/time',
                                          dtype = np.double,
                                          maxshape = (None,),
                                          data = self.data[var]['time'])
                else:
                    h5file.create_dataset(var + '/data',
                                          dtype = self.data[var]['data'].dtype,
                                          maxshape=(None, len(d['uids'])),
                                          data = self.data[var]['data'])
                h5file.create_dataset(var + '/uids',
                                      data = np.array(d['uids'], dtype = 'S'))

if __name__ == '__main__':
    import networkx as nx
    from neurokernel.LPU.LPU import LPU
    from neurokernel.LPU.OutputProcessors.OutputRecorder import OutputRecorder

    dur = 0.2
    dt = 1e-4
    steps = int(dur/dt)
    G = nx.MultiDiGraph()
    uid = 'synapse1'
    scale = 0.1
    ar = 0.110
    ad = 0.190
    gmax = 0.01
    G.add_node(uid, name = 'alpha_synapse',
               model = 'AlphaSynapse2',
               ar = ar,
               ad = ad,
               scale = scale,
               gmax = gmax)

    spike_times = np.array([0.038, 0.048, 0.078], np.float64)
    input_processor = ArrayInputProcessor(
        {'spike_state':
            {'uids': [uid],
             'data': {'time': spike_times,
                      'index': np.array([0]*spike_times.shape[0], np.int32)}
            }
        })
    output_processor = OutputRecorder([('g', [uid])], dur, dt)
    lpu = LPU(dt, 'obj', {'graph': G, 'kwargs': {'class_key': 'model'}},
              device = 0,
              id = 'test', input_processors = [input_processor],
              output_processors = [output_processor],
              debug = True,
              manager = False,
              print_timing = False,
              extra_comps = [ArrayInputProcessor])
    lpu.run(steps = steps)

    s = np.zeros(steps, np.double)
    t = np.arange(0, dur, dt)
    for tk in spike_times:
        s += (scale/(ar-ad) * (np.exp(-ad*(t-tk)*1000) - np.exp(-ar*(t-tk)*1000) ))*(t>tk)

    if np.abs(output_processor.output['g']['data'].reshape(-1)-s).max() < 1e-12:
        print('Test Passed')
    else:
        print('Test Failed')



from neurokernel.LPU.NDComponents.AxonHillockModels.BaseAxonHillockModel import *

class HodgkinHuxley(BaseAxonHillockModel):
    updates = ['spike_state', # (bool)
               'V' # Membrane Potential (mV)
              ]
    accesses = ['I'] # Current (\mu A/cm^2)
    params = ['n', # state variable for activation of K channel ([0-1] unitless)
              'm', # state variable for activation of Na channel ([0-1] unitless)
              'h'  # state variable for inactivation of Na channel ([0-1] unitless)
              ]
    extra_params = ['initV']
    internals = OrderedDict([('V',-65.),  # Membrane Potential (mV)
                             ('Vprev1',-65.)]) # Membrane Potential (mV)

    @property
    def maximum_dt_allowed(self):
        return 1e-5

    def pre_run(self, update_pointers):
        super(HodgkinHuxley, self).pre_run(update_pointers)
        # if 'initV' in self.params_dict:
        self.add_initializer('initV', 'Vprev1', update_pointers)

    def get_update_template(self):
        template = """
#define EXP exp%(fletter)s
#define POW pow%(fletter)s
#define ABS fabs%(fletter)s
#define MIN fmin%(fletter)s

__global__ void update(
    int num_comps,
    %(dt)s dt,
    int nsteps,
    %(input_I)s* g_I,
    %(param_n)s* g_n,
    %(param_m)s* g_m,
    %(param_h)s* g_h,
    %(internal_V)s* g_internalV,
    %(internal_Vprev1)s* g_Vprev1,
    %(update_spike_state)s* g_spike_state,
    %(update_V)s* g_V)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    %(dt)s ddt = dt*1000.; // s to ms

    %(update_V)s V, Vprev1, Vprev2, dV;
    %(input_I)s I;
    %(update_spike_state)s spike;
    %(param_n)s n, dn;
    %(param_m)s m, dm;
    %(param_h)s h, dh;
    %(param_n)s a;

    for(int i = tid; i < num_comps; i += total_threads)
    {
        spike = 0.0;
        V = g_internalV[i];
        Vprev1 = V;
        Vprev2 = g_Vprev1[i];
        I = g_I[i];
        n = g_n[i];
        m = g_m[i];
        h = g_h[i];

        for (int j = 0; j < nsteps; ++j)
        {
            a = exp(-(V+55)/10)-1;
            if (ABS(a) <= 1e-7)
                dn = (1.-n) * 0.1 - n * (0.125*EXP(-(V+65.)/80.));
            else
                dn = (1.-n) * (-0.01*(V+55.)/a) - n * (0.125*EXP(-(V+65)/80));

            a = exp(-(V+40.)/10.)-1.;
            if (ABS(a) <= 1e-7)
                dm = (1.-m) - m*(4*EXP(-(V+65)/18));
            else
                dm = (1.-m) * (-0.1*(V+40.)/a) - m * (4.*EXP(-(V+65.)/18.));

            dh = (1.-h) * (0.07*EXP(-(V+65.)/20.)) - h / (EXP(-(V+35.)/10.)+1.);

            dV = I - 120.*POW(m,3)*h*(V-50.) - 36. * POW(n,4) * (V+77.) - 0.3 * (V+54.387);

            n += ddt * dn;
            m += ddt * dm;
            h += ddt * dh;
            V += ddt * dV;

            spike += (Vprev2<=Vprev1) && (Vprev1 >= V) && (Vprev1 > -30);

            Vprev2 = Vprev1;
            Vprev1 = V;
        }

        g_n[i] = n;
        g_m[i] = m;
        g_h[i] = h;
        g_V[i] = V;
        g_internalV[i] = Vprev1;
        g_Vprev1[i] = Vprev2;
        g_spike_state[i] = MIN(spike, 1.0);
    }
}
"""
        return template


if __name__ == '__main__':
    import argparse
    import itertools
    import networkx as nx
    from neurokernel.tools.logging import setup_logger
    import neurokernel.core_gpu as core

    from neurokernel.LPU.LPU import LPU

    from neurokernel.LPU.InputProcessors.FileInputProcessor import FileInputProcessor
    from neurokernel.LPU.InputProcessors.StepInputProcessor import StepInputProcessor
    from neurokernel.LPU.OutputProcessors.FileOutputProcessor import FileOutputProcessor

    import neurokernel.mpi_relaunch

    dt = 1e-4
    dur = 1.0
    steps = int(dur/dt)

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', default=False,
                        dest='debug', action='store_true',
                        help='Write connectivity structures and inter-LPU routed data in debug folder')
    parser.add_argument('-l', '--log', default='none', type=str,
                        help='Log output to screen [file, screen, both, or none; default:none]')
    parser.add_argument('-s', '--steps', default=steps, type=int,
                        help='Number of steps [default: %s]' % steps)
    parser.add_argument('-g', '--gpu_dev', default=0, type=int,
                        help='GPU device number [default: 0]')
    args = parser.parse_args()

    file_name = None
    screen = False
    if args.log.lower() in ['file', 'both']:
        file_name = 'neurokernel.log'
    if args.log.lower() in ['screen', 'both']:
        screen = True
    logger = setup_logger(file_name=file_name, screen=screen)

    man = core.Manager()

    G = nx.MultiDiGraph()

    G.add_node('neuron0', **{
               'class': 'HodgkinHuxley',
               'name': 'HodgkinHuxley',
               'n': 0.,
               'm': 0.,
               'h': 1.,
               })

    comp_dict, conns = LPU.graph_to_dicts(G)

    fl_input_processor = StepInputProcessor('I', ['neuron0'], 40, 0.2, 0.8)
    fl_output_processor = FileOutputProcessor([('spike_state', None),('V', None)], 'new_output.h5', sample_interval=1)

    man.add(LPU, 'ge', dt, comp_dict, conns,
            device=args.gpu_dev, input_processors = [fl_input_processor],
            output_processors = [fl_output_processor], debug=args.debug)

    man.spawn()
    man.start(steps=args.steps)
    man.wait()

    # plot the result
    import h5py
    import matplotlib
    matplotlib.use('PS')
    import matplotlib.pyplot as plt

    f = h5py.File('new_output.h5')
    t = np.arange(0, args.steps)*dt

    plt.figure()
    plt.plot(t,list(f['V'].values())[0])
    plt.xlabel('time, [s]')
    plt.ylabel('Voltage, [mV]')
    plt.title('Hodgkin-Huxley Neuron')
    plt.savefig('hhn.png',dpi=300)



class ControlInterface(object):
    def __init__(self):
        self.LPU = None
        self.commands = []
    
    def register(self, LPU):
        self.LPU = LPU

    def process_commands(self):
        pass

    # This should be a rpc
    def add_command(self):
        pass



#!/usr/bin/env python


import numpy as np
import pycuda.driver as cuda
from pycuda.compiler import SourceModule
from pycuda.tools import dtype_to_ctype, context_dependent_memoize

"""
This file contains functions to get prepared cuda functions for parray.py.
All functions here should not be used directly.
"""


def _get_type(dtype):
    return dtype.type if isinstance(dtype, np.dtype) else dtype


@context_dependent_memoize
def get_fill_function(dtype, pitch = True):
    type_dst = dtype_to_ctype(dtype)
    name = "fill"
    
    if pitch:
        func = SourceModule(
            fill_pitch_template % {
                    "name": name,
                    "type_dst": type_dst
            }, options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPi'+np.dtype(dtype).char)
        #    [np.int32, np.int32, np.intp, np.int32, _get_type(dtype)])
    else:
        func = SourceModule(
                fill_nonpitch_template % {
                    "name": name,
                    "type_dst": type_dst
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iP'+np.dtype(dtype).char)#[np.int32, np.intp, _get_type(dtype)])
    return func


@context_dependent_memoize
def get_astype_function(dtype_dest, dtype_src, pitch = True):
    type_dest = dtype_to_ctype(dtype_dest)
    type_src = dtype_to_ctype(dtype_src)
    name = "astype"
    operation = ""
    
    if pitch:
        func = SourceModule(
                pitch_template % {
                    "name": name,
                    "dest_type": type_dest,
                    "src_type": type_src,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi')
        #        [np.int32, np.int32, np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_template % {
                    "name": name,
                    "dest_type": type_dest,
                    "src_type": type_src,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPi')#[np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize    
def get_realimag_function(dtype, real = True, pitch = True):
    type_src = dtype_to_ctype(dtype)
    
    if dtype == np.complex64:
        type_dest = "float"
        if real:
            operation = "pycuda::real"
            name = "real"
        else:
            operation = "pycuda::imag"
            name = "imag"
    elif dtype == np.complex128:
        type_dest = "double"
        if real:
            operation = "pycuda::real"
            name = "real"
        else:
            operation = "pycuda::imag"
            name = "imag"
    else:
        raise TypeError("only support complex inputs are "
                        "numpy.complex64 or numpy.complex128")
    
    if pitch:
        func = SourceModule(
                pitch_template % {
                    "name": name,
                    "dest_type": type_dest,
                    "src_type": type_src,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi')
        #        [np.int32, np.int32, np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_template % {
                    "name": name,
                    "dest_type": type_dest,
                    "src_type": type_src,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPi')#[np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_abs_function(dtype, pitch = True):
    type_src = dtype_to_ctype(dtype)
    if dtype == np.complex128:
        operation = "pycuda::abs"
        type_dest = "double"
    elif dtype == np.complex64:
        operation = "pycuda::abs"
        type_dest = "float"
    elif dtype == np.float64:
        operation = "fabs"
        type_dest = "double"
    elif dtype == np.float32:
        operation = "fabsf"
        type_dest = "float"
    else:
        operation = "abs"
        type_dest = dtype_to_ctype(dtype)
    name = "abs_function"
    
    if pitch:
        func = SourceModule(
                pitch_template % {
                    "name": name,
                    "dest_type": type_dest,
                    "src_type": type_src,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi')
        #        [np.int32, np.int32, np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_template % {
                    "name": name,
                    "dest_type": type_dest,
                    "src_type": type_src,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPi')#[np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_angle_function(dtypein, dtypeout, pitch = True):
    type_src = dtype_to_ctype(dtypein)
    type_dest = dtype_to_ctype(dtypeout)
    name = "angle_function"
    if dtypeout == np.float32:
        fletter = "f"
    else:
        fletter = ""
    
    if pitch:
        func = SourceModule(
                pitch_angle_template % {
                    "name": name,
                    "dest_type": type_dest,
                    "src_type": type_src,
                    "fletter": fletter,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi')
        #        [np.int32, np.int32, np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_angle_template % {
                    "name": name,
                    "dest_type": type_dest,
                    "src_type": type_src,
                    "fletter": fletter,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPi')#[np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_conj_function(dtype, pitch = True):
    type_src = dtype_to_ctype(dtype)
    if dtype == np.complex128:
        operation = "pycuda::conj"
    elif dtype == np.complex64:
        operation = "pycuda::conj"
    else:
        raise TypeError("Only complex arrays are allowed "
                        "to perform conjugation")
    name = "conj"
    
    if pitch:
        func = SourceModule(
                pitch_template % {
                    "name": name,
                    "dest_type": type_src,
                    "src_type": type_src,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi')
        #        [np.int32, np.int32, np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_template % {
                    "name": name,
                    "dest_type": type_src,
                    "src_type": type_src,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPi')#[np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_resize_function(dtype):
    type_src = dtype_to_ctype(dtype)
    name = "resize"
    func = SourceModule(
            reshape_template % {
                "name": name,
                "dest_type": type_src,
                "src_type": type_src,
                "operation": "",
            },
            options=["--ptxas-options=-v"]).get_function(name)
    func.prepare('iiiiPiPi')#[np.int32, np.int32, np.int32, np.int32,
    #              np.intp, np.int32, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_transpose_function(dtype, conj = False):
    src_type = dtype_to_ctype(dtype)
    name = "trans"
    operation = ""
    
    if conj:
        if dtype == np.complex128:
            operation = "pycuda::conj"
        elif dtype == np.complex64:
            operation = "pycuda::conj"
    
    func = SourceModule(
            transpose_template % {
                "name": name,
                "type": src_type,
                "operation": operation
            },
            options=["--ptxas-options=-v"]).get_function(name)
    func.prepare('iiPiPi')#[np.int32, np.int32, np.intp,
    #              np.int32, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_addarray_function(left_dtype, right_dtype,
                          rslt_dtype, pitch = True):
    type_left = dtype_to_ctype(left_dtype)
    type_right = dtype_to_ctype(right_dtype)
    type_rslt = dtype_to_ctype(rslt_dtype)

    name = "addarray"
    operation = "+"
    
    if pitch:
        func = SourceModule(
                pitch_array_op_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPiPi')#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_array_op_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPPi')#[np.intp, np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_addscalar_function(src_type, dest_type, pitch = True):
    type_src = dtype_to_ctype(src_type)
    type_dest = dtype_to_ctype(dest_type)
    name = "addscalar"
    operation = "+"
    
    if pitch:
        func = SourceModule(
                pitch_left_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi'+np.dtype(dest_type).char)#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, _get_type(dest_type)])
    else:
        func = SourceModule(
                non_pitch_left_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PP'+np.dtype(dest_type).char+'i')#[np.intp, np.intp, _get_type(dest_type), np.int32])
    return func

@context_dependent_memoize
def get_powarray_function(left_dtype, right_dtype,
                          rslt_dtype, pitch = True):
    type_left = dtype_to_ctype(left_dtype)
    type_right = dtype_to_ctype(right_dtype)
    type_rslt = dtype_to_ctype(rslt_dtype)

    name = "powarray"
    operation = "pow"
    
    if pitch:
        func = SourceModule(
                pitch_array_func_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                    "fletter": 'f' if left_dtype == np.float32 else '',
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPiPi')#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_array_func_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                    "fletter": 'f' if left_dtype == np.float32 else ''
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPPi')#[np.intp, np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_powscalar_function(src_type, dest_type, pitch = True):
    type_src = dtype_to_ctype(src_type)
    type_dest = dtype_to_ctype(dest_type)
    name = "powscalar"
    operation = "pow"
    
    if pitch:
        func = SourceModule(
                pitch_left_scalar_func_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                    "fletter": 'f' if src_type == np.float32 else '',
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi'+np.dtype(dest_type).char)#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, _get_type(dest_type)])
    else:
        func = SourceModule(
                non_pitch_left_scalar_func_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                    "fletter": 'f' if src_type == np.float32 else '',
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PP'+np.dtype(dest_type).char+'i')#[np.intp, np.intp, _get_type(dest_type), np.int32])
    return func


@context_dependent_memoize
def get_subarray_function(left_dtype, right_dtype, rslt_dtype, pitch = True):
    type_left = dtype_to_ctype(left_dtype)
    type_right = dtype_to_ctype(right_dtype)
    type_rslt = dtype_to_ctype(rslt_dtype)
    name = "subdarray"
    operation = "-"
    
    if pitch:
        func = SourceModule(
                pitch_array_op_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPiPi')#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_array_op_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPPi')#[np.intp, np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_subscalar_function(src_type, dest_type, pitch = True):
    type_src = dtype_to_ctype(src_type)
    type_dest = dtype_to_ctype(dest_type)
    
    name = "subscalar"
    operation = "-"
    
    if pitch:
        func = SourceModule(
                pitch_left_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi'+np.dtype(dest_type).char)#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, _get_type(dest_type)])
    else:
        func = SourceModule(
                non_pitch_left_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PP'+np.dtype(dest_type).char+'i')#[np.intp, np.intp, _get_type(dest_type), np.int32])
    return func


@context_dependent_memoize
def get_scalarsub_function(src_type, dest_type, pitch = True):
    type_src = dtype_to_ctype(src_type)
    type_dest = dtype_to_ctype(dest_type)

    name = "scalarsub"
    operation = "-"
    
    if pitch:
        func = SourceModule(
                pitch_right_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi'+np.dtype(dest_type).char)#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, _get_type(dest_type)])
    else:
        func = SourceModule(
                non_pitch_right_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PP'+np.dtype(dest_type).char+'i')#[np.intp, np.intp, _get_type(dest_type), np.int32])
    return func


@context_dependent_memoize
def get_mularray_function(left_dtype, right_dtype, rslt_dtype, pitch = True):
    type_left = dtype_to_ctype(left_dtype)
    type_right = dtype_to_ctype(right_dtype)
    type_rslt = dtype_to_ctype(rslt_dtype)

    name = "mularray"
    operation = "*"

    if pitch:
        func = SourceModule(
                pitch_array_op_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPiPi')#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_array_op_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPPi')#[np.intp, np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_mulscalar_function(src_type, dest_type, pitch = True):
    type_src = dtype_to_ctype(src_type)
    type_dest = dtype_to_ctype(dest_type)
    
    name = "mulscalar"
    operation = "*"
    
    if pitch:
        func = SourceModule(
                pitch_left_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi'+np.dtype(dest_type).char)#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, _get_type(dest_type)])
    else:
        func = SourceModule(
                non_pitch_left_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PP'+np.dtype(dest_type).char+'i')#[np.intp, np.intp, _get_type(dest_type), np.int32])
    return func


@context_dependent_memoize
def get_divarray_function(left_dtype, right_dtype, rslt_dtype, pitch = True):
    type_left = dtype_to_ctype(left_dtype)
    type_right = dtype_to_ctype(right_dtype)
    type_rslt = dtype_to_ctype(rslt_dtype)

    name = "divarray"
    operation = "/"
    
    if pitch:
        func = SourceModule(
                pitch_array_op_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPiPi')#[np.int32, np.int32, np.intp, np.int32,
        #             np.intp, np.int32, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_array_op_template % {
                    "name": name,
                    "dest_type": type_rslt,
                    "left_type": type_left,
                    "right_type": type_right,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPPi')#[np.intp, np.intp, np.intp, np.int32])
    return func


@context_dependent_memoize
def get_divscalar_function(src_type, dest_type, pitch = True):
    type_src = dtype_to_ctype(src_type)
    type_dest = dtype_to_ctype(dest_type)
    
    name = "divscalar"
    operation = "/"
    
    if pitch:
        func = SourceModule(
                pitch_left_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi'+np.dtype(dest_type).char)#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, _get_type(dest_type)])
    else:
        func = SourceModule(
                non_pitch_left_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PP'+np.dtype(dest_type).char+'i')#[np.intp, np.intp, _get_type(dest_type), np.int32])
    return func


@context_dependent_memoize
def get_scalardiv_function(src_type, dest_type, pitch = True):
    type_src = dtype_to_ctype(src_type)
    type_dest = dtype_to_ctype(dest_type)
    
    name = "scalardiv"
    operation = "/"
    
    if pitch:
        func = SourceModule(
                pitch_right_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPi'+np.dtype(dest_type).char)#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.int32, _get_type(dest_type)])
    else:
        func = SourceModule(
                non_pitch_right_scalar_op_template % {
                    "name": name,
                    "src_type": type_src,
                    "dest_type": type_dest,
                    "operation": operation,
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PP'+np.dtype(dest_type).char+'i')#[np.intp, np.intp, _get_type(dest_type), np.int32])
    return func


@context_dependent_memoize
def get_complex_function(real_type, imag_type, result_type, pitch = True):
    type_real = dtype_to_ctype(real_type)
    type_imag = dtype_to_ctype(imag_type)
    type_result = dtype_to_ctype(result_type)
    
    name = "makecomplex"
    
    if pitch:
        func = SourceModule(
                pitch_complex_template % {
                    "name": name,
                    "real_type": type_real,
                    "imag_type": type_imag,
                    "result_type": type_result
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPPi')#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_complex_template % {
                    "name": name,
                    "real_type": type_real,
                    "imag_type": type_imag,
                    "result_type": type_result
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPPi')#[np.intp, np.intp, np.intp, np.int32])
    return func

@context_dependent_memoize
def get_complex_from_amp_function(in_type, result_type, pitch = True):
    type_in = dtype_to_ctype(in_type)
    type_result = dtype_to_ctype(result_type)
    
    name = "makecomplex_amp_phase"
    
    if pitch:
        func = SourceModule(
                pitch_complex_amp_template % {
                    "name": name,
                    "in_type": type_in,
                    "result_type": type_result,
                    "fletter": 'f' if in_type == np.float32 else ''
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('iiPiPPi')#[np.int32, np.int32, np.intp, np.int32,
        #              np.intp, np.intp, np.int32])
    else:
        func = SourceModule(
                non_pitch_complex_amp_template % {
                    "name": name,
                    "in_type": type_in,
                    "result_type": type_result,
                    "fletter": 'f' if in_type == np.float32 else ''
                },
                options=["--ptxas-options=-v"]).get_function(name)
        func.prepare('PPPi')#[np.intp, np.intp, np.intp, np.int32])
    return func


"""templates"""
            
pycuda_complex_header = """
#include <pycuda-complex.hpp>
extern "C++" {
namespace pycuda{

template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator+(
    const complex<_Tp1>& __z1, const complex<_Tp2>& __z2)
{return complex<double>(__z1._M_re + __z2._M_re, __z1._M_im + __z2._M_im);}

template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator-(
    const complex<_Tp1>& __z1, const complex<_Tp2>& __z2)
{return complex<double>(__z1._M_re - __z2._M_re, __z1._M_im - __z2._M_im);}

template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator*(
    const complex<_Tp1>& __z1, const complex<_Tp2>& __z2)
{return complex<double>(\
    __z1._M_re * __z2._M_re - __z1._M_im * __z2._M_im,\
    __z1._M_re * __z2._M_im + __z1._M_im * __z2._M_re);}

template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator/(
    const complex<_Tp1>& __z1, const complex<_Tp2>& __z2)
{
    double c = __z2._M_re;
    double d = __z2._M_im;
    double nom = __z2._M_re*__z2._M_re+__z2._M_im*__z2._M_im;
    return complex<double>((__z1._M_re*__z2._M_re+__z1._M_im*__z2._M_im)/nom,\
                           (__z1._M_im*__z2._M_re-__z1._M_re*__z2._M_im)/nom);
}
    
template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator+(
    const complex<_Tp1>& __z, const _Tp2& __x)
{return complex<double>(__z._M_re + __x, __z._M_im);}

template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator+(
    const _Tp1& __x, const complex<_Tp2>& __z)
{return complex<double>(__z._M_re + __x, __z._M_im);}
    
template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator-(
    const complex<_Tp1>& __z, const _Tp2& __x)
{return complex<double>(__z._M_re - __x, __z._M_im);}

template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator-(
    const _Tp1& __x, const complex<_Tp2>& __z)
{return complex<double>(__x - __z._M_re, -__z._M_im);}
    
template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator*(
    const complex<_Tp1>& __z, const _Tp2& __x)
{return complex<double>(__z._M_re * __x, __z._M_im * __x);}
    
template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator*(
    const _Tp1& __x, const complex<_Tp2>& __z)
{return complex<double>(__z._M_re * __x, __z._M_im * __x);}

template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator/(
    const complex<_Tp1>& __z, const _Tp2& __x)
{return complex<double>(__z._M_re / __x, __z._M_im / __x);}

template <class _Tp1, class _Tp2>
__device__ inline complex<double> operator/(
    const _Tp1& __x, const complex<_Tp2>& __z)
{
    double nom = __z._M_re*__z._M_re+__z._M_im*__z._M_im;
    return complex<double>(__x*__z._M_re/nom, -__x*__z._M_im/nom);
}


}
}

"""



pitch_complex_template = """
#include <pycuda-complex.hpp>

__global__ void
%(name)s(const int M, const int N, %(result_type)s *odata,
         const int ldo, const %(real_type)s *real,
         const %(imag_type)s *imag, const int ldi)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;

    int m, n, idx;
    %(result_type)s tmp;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
    
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            tmp = %(result_type)s(real[m*ldi+idx], imag[m*ldi+idx]);
            odata[m * ldo + idx] = (tmp);
        }
    }
}

"""


non_pitch_complex_template = """
#include <pycuda-complex.hpp>

__global__ void
%(name)s(%(result_type)s *odata, const %(real_type)s *real,
         const %(imag_type)s *imag, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;

    %(result_type)s tmp;

    for (int i = tid; i < N; i += totalthreads)
    {
        tmp = %(result_type)s(real[i], imag[i]);
        odata[i] = (tmp);
    }
}

"""

pitch_complex_amp_template = """
#include <pycuda-complex.hpp>

__global__ void
%(name)s(const int M, const int N, %(result_type)s *odata,
         const int ldo, const %(in_type)s *amp,
         const %(in_type)s *phase, const int ldi)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;

    int m, n, idx;
    %(result_type)s tmp;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
    %(in_type)s c,s, a;
    
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            sincos%(fletter)s(phase[m*ldi+idx], &s, &c);
            a = amp[m*ldi+idx];
            tmp = %(result_type)s(a*c, a*s);
            odata[m * ldo + idx] = (tmp);
        }
    }
}

"""


non_pitch_complex_amp_template = """
#include <pycuda-complex.hpp>

__global__ void
%(name)s(%(result_type)s *odata, const %(in_type)s *amp,
         const %(in_type)s *phase, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;

    %(result_type)s tmp;
    %(in_type)s c,s, a;

    for (int i = tid; i < N; i += totalthreads)
    {
        sincos%(fletter)s(phase[i], &s, &c);
        a = amp[i];
        tmp = %(result_type)s(a*c, a*s);
        odata[i] = (tmp);
    }
}

"""

transpose_template = """
#include <pycuda-complex.hpp>
#define TILE_DIM 32
#define BLOCK_ROWS 8
    
__global__ void
%(name)s(const int M, const int N, %(type)s *odata,
         const int ldo, const %(type)s *idata, const int ldi)
{
    __shared__ %(type)s tile[TILE_DIM][TILE_DIM+1];
    int xIndex, yIndex, index_in, index_out;
    int MM = ((M-1) >> 5) + 1;
    int NN = ((N-1) >> 5) + 1;

    for(int i = blockIdx.x; i < MM * NN; i += gridDim.x)
    {
        xIndex = (i %% NN) * TILE_DIM + threadIdx.x;
        yIndex = (i / NN) * TILE_DIM + threadIdx.y;
        index_in = xIndex + ldi * yIndex;
        if(xIndex < N)
        {
            for(int j = 0; j < TILE_DIM; j+=BLOCK_ROWS)
            {
                if(yIndex + j < M)
                {
                    tile[threadIdx.y+j][threadIdx.x] = \
                        (idata[index_in+j*ldi]);
                }
            }
        }
        __syncthreads();

        xIndex = (i / NN) * TILE_DIM + threadIdx.x;
        yIndex = (i %% NN) * TILE_DIM + threadIdx.y;
        index_out = xIndex + ldo * yIndex;

        if(xIndex < M)
        {
            for(int j = 0; j < TILE_DIM; j+=BLOCK_ROWS)
            {
                if(yIndex + j < N)
                {
                    odata[index_out+j*ldo] = \
                        %(operation)s(tile[threadIdx.x][threadIdx.y+j]);
                }
            }
        }
        __syncthreads();
    }
}

"""

pitch_angle_template = """
#include <pycuda-complex.hpp>

__global__ void
%(name)s(const int M, const int N, %(dest_type)s *dest,
         const int ld_dest, const %(src_type)s *src, const int ld_src)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;

    int m, n, idx;
    %(src_type)s tmp;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            tmp = src[m * ld_src + idx];
            dest[m * ld_dest + idx] = atan2%(fletter)s(pycuda::imag(tmp), pycuda::real(tmp));
        }
    }
}

"""
"""
launching MULTIPROCESSOR_COUNT*6 blocks of (32,8,1),
M: number of rows, N: number of columns,
ld: leading dimension entries(aasumed to be row major)
"""

non_pitch_angle_template = """
#include <pycuda-complex.hpp>
            
__global__ void
%(name)s (%(dest_type)s *dest, const %(src_type)s *src, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    %(src_type)s tmp;
    for (int i = tid; i < N; i += totalthreads)
    {
        tmp = src[i];
        dest[i] =  atan2%(fletter)s(pycuda::imag(tmp), pycuda::real(tmp));
    }
}

"""
""" launching MULTIPROCESSOR_COUNT*6 blocks of (256,1,1), N = totalsize"""


pitch_template = """
#include <pycuda-complex.hpp>

__global__ void
%(name)s(const int M, const int N, %(dest_type)s *dest,
         const int ld_dest, const %(src_type)s *src, const int ld_src)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;

    int m, n, idx;
    %(src_type)s tmp;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            tmp = src[m * ld_src + idx];
            dest[m * ld_dest + idx] = %(operation)s (tmp);
        }
    }
}

"""
"""
launching MULTIPROCESSOR_COUNT*6 blocks of (32,8,1),
M: number of rows, N: number of columns,
ld: leading dimension entries(aasumed to be row major)
"""

            
non_pitch_template = """
#include <pycuda-complex.hpp>
            
__global__ void
%(name)s (%(dest_type)s *dest, const %(src_type)s *src, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    %(src_type)s tmp;
    for (int i = tid; i < N; i += totalthreads)
    {
        tmp = src[i];
        dest[i] =  %(operation)s (tmp);
    }
}

"""
""" launching MULTIPROCESSOR_COUNT*6 blocks of (256,1,1), N = totalsize"""


reshape_template = """
#include <pycuda-complex.hpp>
#include <cuComplex.h>
__global__ void
%(name)s(const int Msrc, const int Nsrc, const int Mdest,
         const int Ndest, %(dest_type)s *dest, const int ld_dest,
         const %(src_type)s *src, const int ld_src)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
    %(src_type)s tmp;
    for (int i = tid; i < Nsrc * Msrc; i += totalthreads)
    {
        tmp = src[i / Nsrc * ld_src + i %% Nsrc];
        dest[i / Ndest * ld_dest + i %% Ndest] = %(operation)s (tmp);
    }
}

"""
"""
launching MULTIPROCESSOR_COUNT*6 blocks of (256,1,1),
M: number of rows, N: number of columns,
ld: leading dimension entries(aasumed to be row major)
"""

irregular_pitch_template = """
#include <pycuda-complex.hpp>

__global__ void
%(name)s(const int M, const int N, %(dest_type)s *dest,
         const int ld_dest, const %(src_type)s *src, const int ld_src)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;
                
    %(src_type)s tmp;
    int m,n;
    for (int i = tid; i < N * M; i += totalthreads)
    {
        m = i / N;
        n = i %% N;
        tmp = src[m * ld_src + n];
        dest[m * ld_dest + n] =  %(operation)s (tmp);
    }
}
    
"""
"""
launching MULTIPROCESSOR_COUNT*6 blocks of (256,1,1),
M: number of rows, N: number of columns, 
ld: leading dimension entries(aasumed to be row major)
"""


pitch_array_op_template = pycuda_complex_header + """
__global__ void
%(name)s(const int M, const int N, %(dest_type)s *dest,
         const int ld_dest, const %(left_type)s *left,
         const int ld_left, const %(right_type)s *right,
         const int ld_right)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;

    int m, n, idx;
    %(left_type)s tmp_left;
    %(right_type)s tmp_right;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
        
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            tmp_left = left[m * ld_left + idx];
            tmp_right = right[m * ld_right + idx];
            dest[m * ld_dest + idx] = (tmp_left) %(operation)s (tmp_right);
        }
    }
}

"""

non_pitch_array_op_template = pycuda_complex_header + """
__global__ void
%(name)s(%(dest_type)s *dest, const %(left_type)s *left,
         const %(right_type)s *right, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;

    %(left_type)s tmp_left;
    %(right_type)s tmp_right;

    for (int i = tid; i < N; i += totalthreads)
    {
        tmp_left = left[i];
        tmp_right = right[i];
        dest[i] =  (tmp_left) %(operation)s (tmp_right);
    }
}

"""

pitch_left_scalar_op_template = pycuda_complex_header + """
__global__ void
%(name)s(const int M, const int N, %(dest_type)s *dest,
         const int ld_dest, const %(src_type)s *left,
         const int ld_left, const %(dest_type)s right)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;
    int m, n, idx;
    %(src_type)s tmp_left;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
    
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            tmp_left = left[m * ld_left + idx];
            dest[m * ld_dest + idx] = (tmp_left) %(operation)s (right);
        }
    }
}

"""

non_pitch_left_scalar_op_template = pycuda_complex_header + """
__global__ void
%(name)s(%(dest_type)s *dest, const %(src_type)s *left,
         const %(dest_type)s right, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;

    %(src_type)s tmp_left;
    for (int i = tid; i < N; i += totalthreads)
    {
        tmp_left = left[i];
        dest[i] =  (tmp_left) %(operation)s (right);
    }
}

"""

pitch_right_scalar_op_template = pycuda_complex_header + """
__global__ void
%(name)s(const int M, const int N, %(dest_type)s *dest,
         const int ld_dest, const %(src_type)s *left,
         const int ld_left, const %(dest_type)s right)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;

    int m, n, idx;
    %(src_type)s tmp_left;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
    
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            tmp_left = left[m * ld_left + idx];
            dest[m * ld_dest + idx] = (right) %(operation)s (tmp_left);
        }
    }
}

"""

non_pitch_right_scalar_op_template = pycuda_complex_header + """
__global__ void
%(name)s(%(dest_type)s *dest, const %(dest_type)s *left,
         const %(src_type)s right, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;

    %(src_type)s tmp_left;
    for (int i = tid; i < N; i += totalthreads)
    {
        tmp_left = left[i];
        dest[i] =  (right) %(operation)s (tmp_left);
    }
}

"""

fill_pitch_template = """
#include <pycuda-complex.hpp>

__global__ void
%(name)s(const int M, const int N, %(type_dst)s *dst,
         const int ld_dst, %(type_dst)s value)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;

    int m, n;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        dst[m * ld_dst + (n<<5) + tid] = value; 
     }
}

"""

"""
launching MULTIPROCESSOR_COUNT*6 blocks of (32,8,1), 
M: number of rows, N: number of columns,
ld: leading dimension entries(aasumed to be row major)
"""

fill_nonpitch_template = """
#include <pycuda-complex.hpp>

__global__ void
%(name)s(const int M,  %(type_dst)s *dst, %(type_dst)s value)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x + blockDim.x * blockIdx.x;
    const int total = gridDim.x * blockDim.x;

    for(int i = tid; i < M; i+=total)
    {
        dst[i] = value; 
    }
}

"""


pitch_array_func_template = pycuda_complex_header + """
__global__ void
%(name)s(const int M, const int N, %(dest_type)s *dest,
         const int ld_dest, const %(left_type)s *left,
         const int ld_left, const %(right_type)s *right,
         const int ld_right)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;

    int m, n, idx;
    %(left_type)s tmp_left;
    %(right_type)s tmp_right;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
        
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            tmp_left = left[m * ld_left + idx];
            tmp_right = right[m * ld_right + idx];
            dest[m * ld_dest + idx] = %(operation)s%(fletter)s((tmp_left) , (tmp_right));
        }
    }
}

"""

non_pitch_array_func_template = pycuda_complex_header + """
__global__ void
%(name)s(%(dest_type)s *dest, const %(left_type)s *left,
         const %(right_type)s *right, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;

    %(left_type)s tmp_left;
    %(right_type)s tmp_right;

    for (int i = tid; i < N; i += totalthreads)
    {
        tmp_left = left[i];
        tmp_right = right[i];
        dest[i] = %(operation)s%(fletter)s((tmp_left) , (tmp_right));
    }
}

"""


pitch_left_scalar_func_template = pycuda_complex_header + """
__global__ void
%(name)s(const int M, const int N, %(dest_type)s *dest,
         const int ld_dest, const %(src_type)s *left,
         const int ld_left, const %(dest_type)s right)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;
    int m, n, idx;
    %(src_type)s tmp_left;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
    
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            tmp_left = left[m * ld_left + idx];
            dest[m * ld_dest + idx] = %(operation)s%(fletter)s((tmp_left) , (right));
        }
    }
}

"""

non_pitch_left_scalar_func_template = pycuda_complex_header + """
__global__ void
%(name)s(%(dest_type)s *dest, const %(src_type)s *left,
         const %(dest_type)s right, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;

    %(src_type)s tmp_left;
    for (int i = tid; i < N; i += totalthreads)
    {
        tmp_left = left[i];
        dest[i] = %(operation)s%(fletter)s((tmp_left) , (right));
    }
}

"""

pitch_right_scalar_func_template = pycuda_complex_header + """
__global__ void
%(name)s(const int M, const int N, %(dest_type)s *dest,
         const int ld_dest, const %(src_type)s *left,
         const int ld_left, const %(dest_type)s right)
{
    //M is the number of rows, N is the number of columns
    const int tid = threadIdx.x;
    const int sid = threadIdx.y + blockDim.y * blockIdx.x;
    const int total = gridDim.x * blockDim.y;

    int m, n, idx;
    %(src_type)s tmp_left;
    int segment_per_row = ((N - 1) >> 5) + 1;
    int total_segments = M * segment_per_row;
    
    for(int i = sid; i < total_segments; i+=total)
    {
        m = i / segment_per_row;
        n = i %% segment_per_row;
        idx = (n<<5) + tid;
        if(idx < N)
        {
            tmp_left = left[m * ld_left + idx];
            dest[m * ld_dest + idx] = %(operation)s%(fletter)s((right) , (tmp_left));
        }
    }
}

"""

non_pitch_right_scalar_func_template = pycuda_complex_header + """
__global__ void
%(name)s(%(dest_type)s *dest, const %(dest_type)s *left,
         const %(src_type)s right, const int N)
{
    const int totalthreads = blockDim.x * gridDim.x;
    const int tid = blockIdx.x * blockDim.x + threadIdx.x;

    %(src_type)s tmp_left;
    for (int i = tid; i < N; i += totalthreads)
    {
        tmp_left = left[i];
        dest[i] = %(operation)s%(fletter)s((right) , (tmp_left));
    }
}

"""



#!/usr/bin/env python

"""
Routines for reading/writing numpy arrays from/to HDF5 files.
"""

import numbers

import h5py
import numpy as np

def dataset_append(dataset, arr):
    """
    Append an array to an h5py dataset.

    Parameters
    ----------
    dataset : h5py.Dataset
        Dataset to extend. Must be resizable in its first dimension.
    arr : numpy.ndarray
        Array to append. All dimensions of `arr` other than the first 
        dimension must be the same as those of the dataset.        
    """
    
    assert isinstance(dataset, h5py.Dataset)
    assert isinstance(arr, np.ndarray)

    # Save leading dimension of stored array:
    maxshape = list(dataset.shape)
    old_ld_dim = maxshape[0]

    # Extend leading dimension of stored array to accommodate new array:
    maxshape[0] += arr.shape[0]
    dataset.resize(maxshape)

    # Compute slices to use when assigning `arr` to array extension:
    slices = [slice(old_ld_dim, None)]
    for s in maxshape[1:]:
        slices.append(slice(None, None))

    # Convert list of slices to tuple because __setitem__ can 
    # only handle "simple" indexes:
    slices = tuple(slices)
    dataset.__setitem__(slices, arr)        

def write_array(arr, filename, mode='w', complevel=0):
    """
    Write numpy array containing numerical data to HDF5 file.

    Parameters
    ----------
    arr : numpy.ndarray, pycuda.gpuarray.GPUArray, parray.PitchArray
        Array to store. Must contain numerical data.
    filename: str
        HDF5 file to write.
    mode : str
        Mode to use when opening file. 'w' creates a new file,
        'a' appends to the dataset in an existing file. If appending, 
        all dimensions of `arr` other than the first dimension must be the same        
        as those of the existing dataset.
    complevel : int
        Compression level. Must be between 0 and 10.

    Notes
    -----
    Files written with this routine can be opened in MATLAB using
    the `h5read` function (although complex values will be returned
    as a structure containing a real array and imaginary array).

    See Also
    --------
    read_array
    """

    if arr.__class__.__name__ in ['GPUArray', 'PitchArray']:
        arr = arr.get()
    elif not isinstance(arr, np.ndarray):
        TypeError('unsupported array type')
    if not issubclass(arr.dtype.type, numbers.Number):
        TypeError('unsupported array dtype')

    h5file = h5py.File(filename, mode)    
    if mode == 'w' or (mode == 'a' and 'array' not in h5file.keys()):
        # Set leading dimension to None to enable the created array to be 
        # resized:
        maxshape = list(arr.shape)
        maxshape[0] = None
        h5file.create_dataset('/array', data=arr, maxshape=maxshape,
                              compression=complevel)
    elif mode == 'a':
        dataset_append(h5file['/array'], arr)
    else:
        RuntimeError('invalid mode')

    h5file.close()

def read_array(filename):
    """
    Read numpy array containing numerical data from HDF5 file.

    Parameters
    ----------
    filename : str
        HDF5 file to read.

    Returns
    -------
    a : numpy.array
        Array read from file.

    Notes
    -----
    Files written with `write_array` or MATLAB's `h5write` function can
    be read with this routine.

    See Also
    --------
    write_array
    """

    h5file = h5py.File(filename, 'r')
    result = h5file['/array'][:]
    h5file.close()
    return result





from collections import OrderedDict

import numpy as np
import h5py
from datetime import datetime
from neurokernel.LPU.OutputProcessors.BaseOutputProcessor import BaseOutputProcessor
import pycuda.driver as cuda
import pycuda.gpuarray as garray


class FileOutputProcessor(BaseOutputProcessor):
    def __init__(self, var_list, filename, sample_interval = 1,
                 cache_length = 1000):
        """
        Parameters
        ----------
        var_list: list
                  a list of 2-tuples, where a 2-tuple has the form ('V', uid_list)
                  to indicate the variable 'V' will be recorded for the
                  components with uids in uid_list.
                  If uid_list is None, i.e., ('V', None)
                  all components with update variable 'V' will be recorded.
        filename: str
                  name of the file to store the data
        sample_interval: int
                         Downsample output data by sample_interval and only
                         store the the downsampled. spike_state will always
                         be stored in index/time fashion (see examples)
                         and will not be affected by sample_interval.
        cache_length: int
                      length (number of steps) of cache on GPU memory to store
                      results before they are flushed to storage drive.

        Examples
        --------
        Recorded data can be retrieved by using the following commands:
        1. For 'spike_state':
            with h5py.File('output.h5') as f:
                index=f['spike_state']['data']['index'][:]
                time=f['spike_state']['data']['time'][:]
                uids=f['spike_state']['uids'][:]
            spikes = {uid.decode(): time[index==i] for i, uid in enumerate(uids)}

        2. For other variables, e.g., 'V':
            with h5py.File('output.h5') as f:
                data=f['V']['data'][:]
                uids=f['V']['uids'][:]
                start_time = f['metadata'].attrs['start_time']
                sample_interval = f['metadata'].attrs['sample_interval']
                dt = f['metadata'].attrs['dt']
            t = np.arange(0,data.shape[0]).reshape((-1,1))*dt*sample_interval+start_time
            V = {uid.decode(): np.hstack((t,data[:,i:i+1])) for i, uid in enumerate(uids)}

        3. meta data can be accessed by:
            with h5py.File('output.h5') as f:
                start_time = f['metadata'].attrs['start_time']
                sample_interval = f['metadata'].attrs['sample_interval']
                dt = f['metadata'].attrs['dt']
                DateCreated  = f['metadata'].attrs['DateCreated']
        """
        self.fname = filename
        self.cache_length = 1000
        super(FileOutputProcessor, self).__init__(var_list, sample_interval)

    def pre_run(self):
        self.h5file = h5py.File(self.fname, 'w')
        self.h5file.create_dataset('metadata', (), 'i')
        self.h5file['metadata'].attrs['start_time'] = self.start_time
        self.h5file['metadata'].attrs['sample_interval'] = self.sample_interval
        self.h5file['metadata'].attrs['dt'] = self.dt
        self.h5file['metadata'].attrs[
            'DateCreated'] = datetime.now().isoformat()

        self.cache = {}
        for var, d in self.variables.items():
            if var == 'spike_state':
                self.h5file.create_dataset(var + '/data/index', (0,), np.int32,
                                           maxshape = (None,))
                self.h5file.create_dataset(var + '/data/time', (0,), np.double,
                                           maxshape = (None,))
                self.h5file.create_dataset(var + '/uids', data=np.array(d['uids'], dtype = 'S'))
                self.cache[var] = garray.empty((self.cache_length, len(d['uids'])),
                                               d['output'].dtype)
            else:
                self.h5file.create_dataset(var + '/data', (0, len(d['uids'])),
                                           d['output'].dtype, # need to be changed later
                                           maxshape=(None, len(d['uids'])))
                self.h5file.create_dataset(var + '/uids', data=np.array(d['uids'], dtype = 'S'))
                self.cache[var] = garray.empty((self.cache_length, len(d['uids'])),
                                               d['output'].dtype)
        self.count = 0
        self.scount = 0
        self.stime_shift = 0

    def get_output_array(self, var):
        if var == 'spike_state':
            return int(self.cache[var].gpudata)+\
                    self.scount*self.cache[var].shape[1]*self.cache[var].dtype.itemsize
        else:
            return int(self.cache[var].gpudata)+\
                   self.count*self.cache[var].shape[1]*self.cache[var].dtype.itemsize

    def process_output(self):
        # for var, d in self.variables.items():
        #     data = self.get_output_gpu(var)
        #     cuda.memcpy_dtod(
        #         int(self.cache[var].gpudata)+\
        #         self.count*self.cache[var].shape[1]*self.cache[var].dtype.itemsize,
        #         data.gpudata, data.nbytes)
        self.count += 1

        if self.count == self.cache_length:
            for var, d in self.variables.items():
                if var != 'spike_state':
                    self.h5file[var + '/data'].resize(
                        (self.h5file[var + '/data'].shape[0] + self.cache_length, len(d['uids'])))
                    self.h5file[var + '/data'][-self.cache_length:, :] = self.cache[var].get()
            self.h5file.flush()
            self.count = 0

    def process_spike_output(self):
        self.scount += 1
        if self.scount == self.cache_length:
            var = 'spike_state'
            t, index = np.where(self.cache[var].get()!=0)
            t = t*self.sim_dt
            nspikes = index.size
            if nspikes:
                self.h5file[var+'/data/time'].resize(
                        (self.h5file[var + '/data/time'].shape[0] + nspikes,))
                self.h5file[var+'/data/index'].resize(
                        (self.h5file[var + '/data/index'].shape[0] + nspikes,))

                self.h5file[var + '/data/time'][-nspikes:] = t+self.stime_shift
                self.h5file[var + '/data/index'][-nspikes:] = index.astype(np.int32)
                self.h5file.flush()
            self.stime_shift += self.cache_length*self.sim_dt
            self.scount = 0

    def post_run(self):
        if self.count > 0:
            for var, d in self.variables.items():
                if var != 'spike_state':
                    self.h5file[var + '/data'].resize(
                        (self.h5file[var + '/data'].shape[0] + self.count, len(d['uids'])))
                    self.h5file[var + '/data'][-self.count:, :] = self.cache[var].get()[:self.count,:]
        if self.scount > 0:
            var = 'spike_state'
            t, index = np.where(self.cache[var].get()[:self.scount,:]!=0)
            t = t * self.sim_dt
            nspikes = index.size
            if nspikes:
                self.h5file[var+'/data/time'].resize(
                        (self.h5file[var + '/data/time'].shape[0] + nspikes,))
                self.h5file[var+'/data/index'].resize(
                        (self.h5file[var + '/data/index'].shape[0] + nspikes,))
                self.h5file[var + '/data/time'][-nspikes:] = t+self.stime_shift
                self.h5file[var + '/data/index'][-nspikes:] = index.astype(np.int32)
        self.h5file.flush()
        self.h5file.close()


class FileOutputReader(object):
    def __init__(self, filename):
        output = {}
        with h5py.File(filename, 'r') as f:
            self.metadata = {'start_time': f['metadata'].attrs['start_time'],
                             'sample_interval': f['metadata'].attrs['sample_interval'],
                             'dt': f['metadata'].attrs['dt'],
                             'DateCreated': f['metadata'].attrs['DateCreated']}
            self.start_time = self.metadata['start_time']
            self.sample_interval = self.metadata['sample_interval']
            self.dt = self.metadata['dt']
            self.DateCreated = self.metadata['DateCreated']

            for var in f:
                if var != 'metadata':
                    if var == 'spike_state':
                        output[var] = {'uids': [n.decode() for n in f[var]['uids'][:]],
                                       'data': {'time': f[var]['data']['time'][:],
                                                'index': f[var]['data']['index'][:]}
                                      }
                    else:
                        output[var] = {'uids': [n.decode() for n in f[var]['uids'][:]],
                                       'data': f[var]['data'][:]}
        self.output = output

    def get_metadata(self):
        return self.metadata

    def get_uids(self, var = None):
        uids = {}
        if var is None:
            for variable in self.output:
                if variable != 'metadata':
                    uids[variable] = self._get_uid_by_var(variable)
        else:
            return self._get_uid_by_var(var)

    def _get_uids_by_var(self, var):
        return self.output[var]['uids']

    def get_output(self, var = None, uids = None):
        """
        retrieve outputs by variable name and uids

        Parameters
        ----------
        var: str
             Name of the variable to retrieve.
             If None, all variables associated with a uid will be retrieved.
        uids: str or list of str
              uids of the component to retrieve
              If not specified (None), all uids associated with the `var`
              will be retrieved.

        Returns
        -------
        output: dict or OrderedDict
                If uids is a list or tuple, returns a OrderedDict, Otherwise a dict,
                with output keyed by uid, and values are either the data/spike_times
                of the component, or a dict keyed by variable name and data/spike_times
                in the value.
        """
        if var is None and uids is None:
            return self.output
        elif var is not None and uids is None:
            return self._get_output_by_var(var)
        elif var is None and uids is not None:
            return self._get_output_by_uids(uids)
        else:
            return self._get_output_by_var_and_uids(var, uids)

    def _get_output_by_var(self, var):
        uids = self.output[var]['uids']
        data = self.output[var]['data']
        if var == 'spike_state':
            output = {uid: {'data': data['time'][data['index']==i] + self.start_time} for i, uid in enumerate(uids)}
        else:
            t = np.arange(0, data.shape[0])*self.dt*self.sample_interval + self.start_time
            output = {uid: {'time': t, 'data': data[:,i].copy()} for i, uid in enumerate(uids)}
        return output

    def _get_output_by_uid(self, uid):
        output = {}
        for var in self.output:
            try:
                index = self.output[var]['uids'].index(uid)
            except ValueError:
                pass
            else:
                if var == 'spike_state':
                    output[var] = {'data': self.output[var]['data']['time'][self.output[var]['data']['index'] == index] + self.start_time}
                else:
                    t = np.arange(0, self.output[var]['data'].shape[0])*self.dt*self.sample_interval + self.start_time
                    output[var] = {'time': t, 'data': self.output[var]['data'][:,index].copy()}
        return output

    def _get_output_by_uids(self, uids):
        if isinstance(uids, str):
            return {uids: self._get_output_by_uid(uids)}
        elif isinstance(uids, (list, tuple)):
            return OrderedDict([(uid, self._get_output_by_uid(uid)) for uid in uids])
        elif isinstance(uids, set):
            return {uid: self._get_output_by_uid(uid) for uid in uids}

    def _get_output_by_var_and_uid(self, var, uid):
        output = None
        try:
            index = self.output[var]['uids'].index(uid)
        except ValueError:
            pass
        else:
            if var == 'spike_state':
                output = {'data': self.output[var]['data']['time'][self.output[var]['data']['index'] == index] + self.start_time}
            else:
                t = np.arange(0, self.output[var]['data'].shape[0])*self.dt*self.sample_interval + self.start_time
                output = {'time': t, 'data': self.output[var]['data'][:,index].copy()}
        return output

    def _get_output_by_var_and_uids(self, var, uids):
        if isinstance(uids, str):
            return {uids: self._get_output_by_var_and_uid(var, uids)}
        elif isinstance(uids, (list, tuple)):
            return OrderedDict([(uid, self._get_output_by_var_and_uid(var, uid)) for uid in uids])
        elif isinstance(uids, set):
            return {uid: self._get_output_by_var_and_uid(var, uid) for uid in uids}



#!/usr/bin/env python

from abc import ABCMeta, abstractmethod, abstractproperty
from future.utils import with_metaclass
from neurokernel.LPU.NDComponents.NDComponent import NDComponent

class BaseDendriteModel(with_metaclass(ABCMeta, NDComponent,object)):
    # __metaclass__ = ABCMeta

    accesses = ['g']
    updates = ['I']




from collections import OrderedDict
import uuid

import h5py
import numpy as np
import networkx as nx

import pycuda.driver as cuda
import pycuda.gpuarray as garray
from neurokernel.LPU.InputProcessors.ArrayInputProcessor import ArrayInputProcessor

class PresynapticInputProcessor(ArrayInputProcessor):
    """
    Allow users to specify inputs to neurons from presynaptic variables such as
    voltage and spike_state, and automatically add user-defined synapses.

    Parameters
    ----------
    input: dict
           A dictionary with the following structure:
           {var1: {'uids': ['comp1', 'comp2' ... ],
                   'data': np.ndarray,
                   'components': {'model': 'AlphaSynpase', 'gmax': 10.0,
                                  'ar': 0.19, 'ad': 0.16}
                  },
            var2: {'uids': ['comp3', 'comp4' ... ],
                   'data': np.ndarray,
                   'components': [{'model': 'SynapseAMPA', 'gmax': 10.0, 'st': 4.0},
                                  {'model': 'SynapseGABA', 'gmax': 8.0, 'st': 4.0},
                                  ...]
                  },
            ...}
           where in each key/value pair, the key specifies the varialbe to be
           inject input to, value['uids'] is a list that specifies the uids of
           components to be injected to, and value['data'] is an np.ndarray
           of shape (T, N) that specifies the input to each of the N components
           at every time step for a total of T steps, and the ordering the data in
           each column should be the same as the order in value['uids'].
           For 'spike_state', in addition to provide 0 (no spike) or 1 (spike)
           for each time step, spike time and the index of neuron that spiked
           can be provided by value['data']['time'] and value['data']['index'],
           respectively. See examples for more details.

           The input can also be programmatically loaded by other methods.
    mode:     int
              mode in BaseInputProcessor
              0: default, when the file does not cover the entire simulation,
                 inputs default to 0.
              1: when the file does not cover the entire simulation,
                 inputs defaults to the last state.
    cache_length: int
                  length (number of steps) of cache on GPU memory to preload
                  inputs.

    Examples
    --------

    """
    def add_inputs(self, inputs):
        for var, g in inputs.items():
            self.add_input(var, g['uids'], g['data'], g['components'])

    def add_input(self, var, uids, data, components):
        new_uids = []
        for i, uid in enumerate(uids):
            if not uid in self._additional_graph.nodes:
                self._additional_graph.add_node(uid)
            new_uid = 'synaptic_input_to_{}_{}'.format(uid, str(uuid.uuid4()).split('-')[0])
            params = components if isinstance(components, dict) else components[i]
            self._additional_graph.add_node(new_uid, **params)
            new_uids.append(new_uid)
            self._additional_graph.add_edge(new_uid, uid)

        if var in self.variables:
            if var == 'spike_state':
                if isinstance(data, dict):
                    assert issubclass(data['index'].dtype.type, np.integer), \
                           'spike index array must be integer dtype'
                    self.data[var]['time'] = np.concatenate(
                            self.data[var]['time'], data['time'])
                    self.data[var]['index'] = np.concatenate(
                            self.data[var]['index'],
                            data['index']+len(self.variables[var]['uids']))
                else:
                    self.data[var] = np.hstack((self.data[var], data))
            else:
                self.data[var] = np.hstack((self.data[var], data))
        else:
            self.data[var] = data
            if var == 'spike_state':
                if isinstance(data, dict):
                    self.spike_state_format = 'event'
                else:
                    self.spike_state_format = 'stepwise'
        self.add_variables([(var, new_uids)])

if __name__ == '__main__':
    import networkx as nx
    from neurokernel.LPU.LPU import LPU
    from neurokernel.LPU.OutputProcessors.OutputRecorder import OutputRecorder

    dur = 0.2
    dt = 1e-4
    steps = int(dur/dt)
    G = nx.MultiDiGraph()
    uid = 'synapse1'
    scale = 0.1
    ar = 0.110
    ad = 0.190
    gmax = 0.01
    G.add_node(uid, name = 'alpha_synapse',
               model = 'AlphaSynapse2',
               ar = ar,
               ad = ad,
               scale = scale,
               gmax = gmax)

    spike_times = np.array([0.038, 0.048, 0.078], np.float64)
    input_processor = ArrayInputProcessor(
        {'spike_state':
            {'uids': [uid],
             'data': {'time': spike_times,
                      'index': np.array([0]*spike_times.shape[0], np.int32)}
            }
        })
    output_processor = OutputRecorder([('g', [uid])], dur, dt)
    lpu = LPU(dt, 'obj', {'graph': G, 'kwargs': {'class_key': 'model'}},
              device = 0,
              id = 'test', input_processors = [input_processor],
              output_processors = [output_processor],
              debug = True,
              manager = False,
              print_timing = False,
              extra_comps = [ArrayInputProcessor])
    lpu.run(steps = steps)

    s = np.zeros(steps, np.double)
    t = np.arange(0, dur, dt)
    for tk in spike_times:
        s += (scale/(ar-ad) * (np.exp(-ad*(t-tk)*1000) - np.exp(-ar*(t-tk)*1000) ))*(t>tk)

    if np.abs(output_processor.output['g']['data'].reshape(-1)-s).max() < 1e-12:
        print('Test Passed')
    else:
        print('Test Failed')



from collections import OrderedDict


import numpy as np
import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype
import pycuda.driver as cuda
from pycuda.compiler import SourceModule

from .BaseSynapseModel import BaseSynapseModel

# assumes a maximum of one input connection per synapse
class SynapseNMDA(BaseSynapseModel):
    accesses = ['spike_state', 'V']
    updates = ['g']
    # params = ['gmax', 'st', 'xt', 'Mg']
    params = ['gmax', 'st', 'Mg']
    internals = OrderedDict([('tst', 0.)])#, ('txt', 0.)])

    @property
    def maximum_dt_allowed(self):
        return 1e-4

    def get_update_template(self):
        template = """
__global__ void update(int num_comps, %(dt)s dt, int nsteps,
                       %(input_spike_state)s* g_spike_state, %(input_V)s* g_V,
                       %(param_gmax)s *g_gmax, %(param_st)s *g_st,
                       %(param_Mg)s *g_Mg,
                       %(internal_tst)s *g_tst,
                       %(update_g)s *g_g)
{
    int tid = threadIdx.x + blockIdx.x*blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    %(dt)s ddt = dt*1000.;
    %(input_spike_state)s spike_state;
    %(param_st)s st, est;
    %(internal_tst)s tst;

    for(int i = tid; i < num_comps; i += total_threads)
    {
        spike_state = g_spike_state[i];
        st = g_st[i];
        tst = g_tst[i];

        est = exp%(fletter)s(-ddt/st);
        for(int k = 0; k < nsteps; ++k)
        {
            if(k == 0 && (spike_state>0))
            {
                tst = fmin%(fletter)s(tst + 0.63*(1-tst), 1.0);
            }
            tst *= est;
        }
        g_g[i] = g_gmax[i]*tst/(1+g_Mg[i]*exp%(fletter)s(-0.062*g_V[i])/3.57);
        g_tst[i] = tst;
    }
}
"""
        return template

#     def get_update_template(self):
#         template = """
# __global__ void update(int num_comps, %(dt)s dt, int nsteps,
#                        %(input_spike_state)s* g_spike_state, %(input_V)s* g_V,
#                        %(param_gmax)s *g_gmax, %(param_st)s *g_st,
#                        %(param_xt)s *g_xt, %(param_Mg)s *g_Mg,
#                        %(internal_tst)s *g_tst, %(internal_txt)s *g_txt,
#                        %(update_g)s *g_g)
# {
#     int tid = threadIdx.x + blockIdx.x*blockDim.x;
#     int total_threads = gridDim.x * blockDim.x;
#
#     %(dt)s ddt = dt*1000.;
#     %(input_spike_state)s spike_state;
#     %(param_st)s st;
#     %(internal_tst)s tst;
#     %(param_xt)s xt, ext;
#     %(internal_txt)s txt;
#
#     for(int i = tid; i < num_comps; i += total_threads)
#     {
#         spike_state = g_spike_state[i];
#         st = g_st[i];
#         tst = g_tst[i];
#         xt = g_xt[i];
#         txt = g_txt[i];
#
#         if(spike_state)
#         {
#             txt += 1;
#         }
#
#         ext = exp%(fletter)s(-ddt/xt);
#         for(int k = 0; k < nsteps; ++k)
#         {
#             txt *= ext;
#             if(k == 0 && (spike_state>0))
#             {
#                 txt += 1;
#             }
#             tst += ddt*( 0.63*txt*(1 - tst) - tst/st);
#         }
#
#         g_g[i] = g_gmax[i]*tst/(1+g_Mg[i]*exp%(fletter)s(-0.062*g_V[i])/3.57);
#         g_tst[i] = tst;
#         g_txt[i] = txt;
#     }
# }
# """
#         return template


if __name__ == '__main__':
    import argparse
    import itertools
    import networkx as nx
    from neurokernel.tools.logging import setup_logger
    import neurokernel.core_gpu as core

    from neurokernel.LPU.LPU import LPU

    from neurokernel.LPU.InputProcessors.FileInputProcessor import FileInputProcessor
    from neurokernel.LPU.InputProcessors.StepInputProcessor import StepInputProcessor
    from neurokernel.LPU.OutputProcessors.FileOutputProcessor import FileOutputProcessor

    import neurokernel.mpi_relaunch

    dt = 1e-4
    dur = 1.0
    steps = int(dur/dt)

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', default=False,
                        dest='debug', action='store_true',
                        help='Write connectivity structures and inter-LPU routed data in debug folder')
    parser.add_argument('-l', '--log', default='none', type=str,
                        help='Log output to screen [file, screen, both, or none; default:none]')
    parser.add_argument('-s', '--steps', default=steps, type=int,
                        help='Number of steps [default: %s]' % steps)
    parser.add_argument('-g', '--gpu_dev', default=0, type=int,
                        help='GPU device number [default: 0]')
    args = parser.parse_args()

    file_name = None
    screen = False
    if args.log.lower() in ['file', 'both']:
        file_name = 'neurokernel.log'
    if args.log.lower() in ['screen', 'both']:
        screen = True
    logger = setup_logger(file_name=file_name, screen=screen)

    man = core.Manager()

    G = nx.MultiDiGraph()

    G.add_node('neuron0', {
               'class': 'LIF',
               'name': 'LIF',
               'resting_potential': -55.0,
               'threshold': -45.0,
               'reset_voltage': -55.0,
               'capacitance': 0.0744005237682, # in mS
               'refractory_period': 3.0, # in milliseconds
               'time_constant': 16.0, # in milliseconds
               'bias_current': 0.0
               })
    G.add_node('synapse0', {
               'class': 'SynapseNMDA',
               'name': 'SynapseNMDA',
               'gmax': 0.0006,#0.43576,
               'st': 100.0,
               'xt': 5.0,
               'reverse': 0.0
               })
    G.add_edge('synapse0', 'neuron0')
    G.add_edge('neuron0', 'synapse0', variable='V')

    comp_dict, conns = LPU.graph_to_dicts(G)

    #fl_input_processor = StepInputProcessor('I', ['neuron0'], 1, 0.2, 0.4)
    fl_input_processor = FileInputProcessor('input_spike.h5')
    fl_output_processor = FileOutputProcessor([('g', None),('spike_state', None),('I', None), ('V', None)], 'new_output.h5', sample_interval=1)

    man.add(LPU, 'ge', dt, comp_dict, conns,
            device=args.gpu_dev, input_processors = [fl_input_processor],
            output_processors = [fl_output_processor], debug=args.debug)

    man.spawn()
    man.start(steps=args.steps)
    man.wait()



from neurokernel.LPU.NDComponents.SynapseModels.BaseSynapseModel import *

class AlphaSynapse(BaseSynapseModel):
    accesses = ['spike_state'] # (bool)
    updates = ['g'] # conductance (mS/cm^2)
    params = ['gmax', # maximum conductance (mS/cm^2)
              'ar', # rise rate of conductance (ms^{-1})
              'ad', # decay rate of conductance (ms^{-1})
              ]
    internals = OrderedDict([('z', 0.0),  # g,
                             ('dz', 0.0),  # derivative of g
                             ('d2z', 0.0)  # second derivative of g
                             ])

    @property
    def maximum_dt_allowed(self):
        return 1e-4

    def get_update_template(self):
        # The following kernel assumes a maximum of one input connection
        # per neuron
        if self.internal_steps == 1:
            # this is a kernel that runs 1 step internally for each self.dt
            template = """
__global__ void update(int num_comps, %(dt)s dt, int nsteps,
                       %(input_spike_state)s* g_spike_state,
                       %(param_gmax)s* g_gmax, %(param_ar)s* g_ar,
                       %(param_ad)s* g_ad,
                       %(internal_z)s* g_z, %(internal_dz)s* g_dz,
                       %(internal_d2z)s* g_d2z, %(update_g)s* g_g)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    %(dt)s ddt = dt*1000.; // s to ms
    %(input_spike_state)s spike_state;
    %(param_gmax)s gmax;
    %(param_ar)s ar;
    %(param_ad)s ad;
    %(internal_z)s z, new_z;
    %(internal_dz)s dz, new_dz;
    %(internal_d2z)s d2z, new_d2z;

    for(int i = tid; i < num_comps; i += total_threads)
    {
        ar = g_ar[i];
        ad = g_ad[i];
        z = g_z[i];
        dz = g_dz[i];
        d2z = g_d2z[i];
        spike_state = g_spike_state[i];

        new_z = fmax( 0., z + ddt*dz );
        new_dz = dz + ddt*d2z;
        if( spike_state>0.0 )
            new_dz += ar*ad;
        new_d2z = -( ar+ad )*dz - ar*ad*z;

        gmax = g_gmax[i];
        g_z[i] = new_z;
        g_dz[i] = new_dz;
        g_d2z[i] = new_d2z;
        g_g[i] = new_z*gmax;
    }
}
"""
        else:
            # this is a kernel that runs self.nstep steps internally for each self.dt
            # see the "k" for loop
            template = """
__global__ void update(int num_comps, %(dt)s dt, int nsteps,
                       %(input_spike_state)s* g_spike_state,
                       %(param_gmax)s* g_gmax, %(param_ar)s* g_ar,
                       %(param_ad)s* g_ad,
                       %(internal_z)s* g_z, %(internal_dz)s* g_dz,
                       %(internal_d2z)s* g_d2z, %(update_g)s* g_g)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    %(dt)s ddt = dt*1000.; // s to ms
    %(input_spike_state)s spike_state;
    %(param_gmax)s gmax;
    %(param_ar)s ar;
    %(param_ad)s ad;
    %(internal_z)s z, new_z;
    %(internal_dz)s dz, new_dz;
    %(internal_d2z)s d2z, new_d2z;

    for(int i = tid; i < num_comps; i += total_threads)
    {
        ar = g_ar[i];
        ad = g_ad[i];
        z = g_z[i];
        dz = g_dz[i];
        d2z = g_d2z[i];
        spike_state = g_spike_state[i];

        for(int k = 0; k < nsteps; ++k)
        {
            new_z = fmax( 0., z + ddt*dz );
            new_dz = dz + ddt*d2z;
            if(k == 0 && (spike_state>0.0))
                new_dz += ar*ad;
            new_d2z = -( ar+ad )*dz - ar*ad*z;

            z = new_z;
            dz = new_dz;
            d2z = new_d2z;
        }

        gmax = g_gmax[i];
        g_z[i] = new_z;
        g_dz[i] = new_dz;
        g_d2z[i] = new_d2z;
        g_g[i] = new_z*gmax;
    }
}
"""
        return template

if __name__ == '__main__':
    import argparse
    import itertools

    import networkx as nx
    import h5py

    from neurokernel.tools.logging import setup_logger
    import neurokernel.core_gpu as core
    from neurokernel.LPU.LPU import LPU
    from neurokernel.LPU.InputProcessors.FileInputProcessor import FileInputProcessor
    from neurokernel.LPU.OutputProcessors.FileOutputProcessor import FileOutputProcessor
    import neurokernel.mpi_relaunch

    dt = 1e-4
    dur = 1.0
    steps = int(dur / dt)

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', default=False,
                        dest='debug', action='store_true',
                        help='Write connectivity structures and inter-LPU routed data in debug folder')
    parser.add_argument('-l', '--log', default='both', type=str,
                        help='Log output to screen [file, screen, both, or none; default:none]')
    parser.add_argument('-s', '--steps', default=steps, type=int,
                        help='Number of steps [default: %s]' % steps)
    parser.add_argument('-g', '--gpu_dev', default=0, type=int,
                        help='GPU device number [default: 0]')
    args = parser.parse_args()

    file_name = None
    screen = False
    if args.log.lower() in ['file', 'both']:
        file_name = 'neurokernel.log'
    if args.log.lower() in ['screen', 'both']:
        screen = True
    logger = setup_logger(file_name=file_name, screen=screen)

    t = np.arange(0, dt * steps, dt)

    uids = np.array(["synapse0"], dtype='S')

    spike_state = np.zeros((steps, 1), dtype=np.float64)
    spike_state[np.nonzero((t - np.round(t / 0.04) * 0.04) == 0)[0]] = 1

    with h5py.File('input_spike.h5', 'w') as f:
        f.create_dataset('spike_state/uids', data=uids)
        f.create_dataset('spike_state/data', (steps, 1),
                         dtype=np.float64,
                         data=spike_state)

    man = core.Manager()

    G = nx.MultiDiGraph()

    G.add_node('synapse0', **{
               'class': 'AlphaSynapse',
               'name': 'AlphaSynapse',
               'gmax': 0.003,
               'ar': 0.11,
               'ad': 0.19,
               'reverse': 0.0
               })

    comp_dict, conns = LPU.graph_to_dicts(G)

    fl_input_processor = FileInputProcessor('input_spike.h5')
    fl_output_processor = FileOutputProcessor(
        [('g', None)], 'new_output.h5', sample_interval=1)

    man.add(LPU, 'ge', dt, comp_dict, conns,
            device=args.gpu_dev, input_processors=[fl_input_processor],
            output_processors=[fl_output_processor], debug=args.debug)

    man.spawn()
    man.start(steps=args.steps)
    man.wait()
